import datetime
import email.utils
import os
import sys
import time
from pathlib import Path

import requests


def download_file(
    url: str,
    outdir: Path,
    retries: int = 10,
    *,
    revalidate: bool = False,
) -> Path:
    """Downloads a file from a URL with cache and retry logic.

    Args:
        url: The URL from which to download the file.
        outdir: The local filesystem path where the downloaded file will be saved.
        retries: Maximum number of retry attempts on failure.
        revalidate: If True, the cached file is revalidated based on
            modification time. Defaults to False.

    Raises:
        requests.HTTPError: If the download fails after all retries.
    """
    filename = url.rsplit("/", maxsplit=1)[-1]
    out = outdir / filename
    attempt = 0
    while True:
        try:
            _download_file(url, out, revalidate=revalidate)
            break
        except requests.HTTPError as e:
            print(
                f"Failed to download file on attempt {attempt + 1}: {e}",
                file=sys.stderr,
            )
            out.unlink(missing_ok=True)
            if attempt >= retries:
                raise
            time.sleep(2**attempt)
        attempt += 1
    return out


def _download_file(url: str, out: Path, *, revalidate: bool) -> None:
    try:
        pending_output = False
        print_progress = sys.stdout.isatty()
        if not print_progress:
            print(f"Download {url}", file=sys.stderr)
        headers = {}
        if out.exists():
            if not revalidate:
                return
            mtime = os.path.getmtime(out)
            headers["If-Modified-Since"] = email.utils.formatdate(mtime, usegmt=True)
        with requests.get(url, headers=headers, stream=True, timeout=60) as res:
            res.raise_for_status()
            # GDAS1 redirects missing files to a page that returns 200.
            if res.url.endswith("/notfound.php"):
                msg = "Page not found"
                raise RuntimeError(msg)
            if res.status_code == 304:
                return
            total_bytes = res.headers.get("Content-Length")
            with out.open("wb") as f:
                if total_bytes is None:
                    f.write(res.content)
                else:
                    dl_bytes = 0
                    total_bytes_int = int(total_bytes)
                    for data in res.iter_content(chunk_size=4096):
                        dl_bytes += len(data)
                        f.write(data)
                        if print_progress:
                            percent = round(100 * dl_bytes / total_bytes_int)
                            print(
                                f"\r[{percent:3}%] {url}",
                                end="",
                                file=sys.stderr,
                                flush=True,
                            )
                            pending_output = True
            if "Last-Modified" in res.headers:
                last_modified = res.headers["Last-Modified"]
                try:
                    new_mtime = email.utils.parsedate_to_datetime(
                        last_modified,
                    ).timestamp()
                    new_atime = datetime.datetime.now().timestamp()
                    os.utime(out, (new_atime, new_mtime))
                except ValueError:
                    pass
    finally:
        if pending_output:
            print(file=sys.stderr)



from .arome import read_arome as read_arome
from .arpege import read_arpege as read_arpege
from .ecmwf_open import read_ecmwf_open as read_ecmwf_open
from .gdas1 import read_gdas1 as read_gdas1
from .icon_d2 import read_icon_d2 as read_icon_d2



import datetime
from pathlib import Path

import netCDF4
from numpy.testing import assert_allclose, assert_array_equal

from model_munger.extract import Extractor, FixedLocation, MobileLocation, write_netcdf
from model_munger.extractors.ecmwf_open import read_ecmwf
from model_munger.readers.ecmwf_open import ECMWF_OPEN


def test_extract_profiles(tmp_path: Path) -> None:
    input_files = [
        "tests/data/20250115000000-0h-oper-fc.grib2",
        "tests/data/20250115000000-3h-oper-fc.grib2",
    ]
    locations: list[FixedLocation | MobileLocation] = [
        FixedLocation(
            id="hyytiala",
            name="Hyytiälä",
            latitude=61.844,
            longitude=24.287,
        ),
        MobileLocation(
            id="boaty",
            name="Boaty McBoatface",
            time=[
                datetime.datetime(2025, 1, 14, 23, 59, tzinfo=datetime.timezone.utc),
                datetime.datetime(2025, 1, 15, 3, 1, tzinfo=datetime.timezone.utc),
            ],
            latitude=[59.446, 59.801],
            longitude=[24.772, 24.839],
        ),
    ]
    time = [
        datetime.datetime(2025, 1, 15, 0, tzinfo=datetime.timezone.utc),
        datetime.datetime(2025, 1, 15, 3, tzinfo=datetime.timezone.utc),
    ]
    extractor = Extractor(time, locations, ECMWF_OPEN)
    for input_file in input_files:
        for level in read_ecmwf(input_file):
            extractor.add_level(level)
    for raw in extractor.extract_profiles():
        write_netcdf(raw, tmp_path / f"{raw.location.id}.nc")
    with netCDF4.Dataset(tmp_path / "hyytiala.nc") as nc:
        assert nc["time"].units == "hours since 2025-01-15 00:00:00 +00:00"
        assert_array_equal(nc["time"][:], [0, 3])
        assert_array_equal(nc["latitude"][:], 61.75)
        assert_array_equal(nc["longitude"][:], 24.25)
        assert_array_equal(nc["pressure"][:], [100_000, 10_000])
        assert_allclose(
            nc["t"][:],
            [[273.766754, 214.335098], [271.976425, 213.248489]],
        )
        assert_allclose(nc["t2m"][:], [273.366211, 270.937653])
        assert_allclose(
            nc["sot"][:],
            [[273.004318, 272.756409], [272.97612, 272.759644]],
        )
    with netCDF4.Dataset(tmp_path / "boaty.nc") as nc:
        assert nc["time"].units == "hours since 2025-01-15 00:00:00 +00:00"
        assert_array_equal(nc["time"][:], [0, 3])
        assert_array_equal(nc["latitude"][:], [59.5, 59.75])
        assert_array_equal(nc["longitude"][:], [24.75, 24.75])
        assert_array_equal(nc["pressure"][:], [100_000, 10_000])
        assert_allclose(
            nc["t"][:],
            [[275.423004, 213.647598], [275.007675, 213.435989]],
        )
        assert_allclose(nc["t2m"][:], [275.584961, 276.343903])
        assert_allclose(
            nc["sot"][:],
            [[274.848068, 274.912659], [277.22612, 277.228394]],
        )



from os import PathLike

import netCDF4
import numpy as np
from cftime import num2pydate

from model_munger.model import Location, Model, ModelType
from model_munger.utils import calc_geometric_height

keymap = {
    "PRSS": "sfc_pressure",
    "MSLP": "sfc_pressure_amsl",
    "TPP6": "sfc_total_rain",
    "UMOF": "sfc_turb_mom_v",
    "VMOF": "sfc_turb_mom_u",
    "SHTF": "sfc_down_sens_heat_flx",
    "DSWF": "sfc_net_sw",
    "RH2M": "sfc_rh_2m",
    "U10M": "sfc_wind_u_10m",
    "V10M": "sfc_wind_v_10m",
    "T02M": "sfc_temp_2m",
    "TCLD": "sfc_cloud_fraction",
    "CAPE": "sfc_cape",
    "CINH": "sfc_cin",
    "LISD": "standard_lifted_index",
    "LIB4": "best_4layer_lifted_index",
    "PBLH": "sfc_bl_height",
    "TMPS": "sfc_temp",
    "CPP6": "sfc_conv_rain",
    "CPPA": "sfc_conv_rain",
    "SOLM": "sfc_soil_moisture",
    "CSNO": "sfc_categorical_snow",
    "CICE": "sfc_categorical_ice",
    "CFZR": "sfc_categorical_freezing_rain",
    "CRAI": "sfc_categorical_rain",
    "LHTF": "sfc_net_lat_heat_flx",
    "LCLD": "sfc_cloud_fraction_low",
    "MCLD": "sfc_cloud_fraction_medium",
    "HCLD": "sfc_cloud_fraction_high",
    "TEMP": "temperature",
    "UWND": "uwind",
    "VWND": "vwind",
    "WWND": "omega",
    "RELH": "rh",
    "pressure": "pressure",
    "forecast_time": "forecast_time",
    "latitude": "latitude",
    "longitude": "longitude",
    "horizontal_resolution": "horizontal_resolution",
}

units_map = {
    "J/kg": "J kg-1",
    "N/m2": "kg m-1 s-2",
    "W/m2": "W m-2",
    "frac.": "1",
    "hPa/s": "hPa s-1",
    "m/s": "m s-1",
}


RH_COMMENT = (
    "With respect to water for temperatures greater than or equal to zero\n"
    "degrees Celsius (0C), with respect to ice for temperatures less than\n"
    "-20C, and a blend between -20C and 0C"
)

WATER_DENSITY = 1000
"Approximate density of water (kg m-3)"


def read_gdas1(file: str | PathLike, location: Location) -> Model:
    """Read GDAS1 netCDF generated using model-munger."""
    with netCDF4.Dataset(file) as nc:
        data = {}
        units = {}
        sources = {}
        comments = {"rh": RH_COMMENT, "sfc_rh_2m": RH_COMMENT}

        for src, dst in keymap.items():
            if src not in nc.variables:
                continue
            ncvar = nc[src]
            values = ncvar[:]
            if src in ("TPP6", "CPP6", "CPPA") and ncvar.units == "m":
                data[dst] = values * WATER_DENSITY
                units[dst] = "kg m-2"
                sources[dst] = f"{src} converted from {ncvar.units} to {units[dst]}"
            elif src in ("CSNO", "CICE", "CFZR", "CRAI"):
                data[dst] = values.astype(np.int16)
                units[dst] = "1"
                sources[dst] = src
            else:
                data[dst] = values
                units[dst] = units_map.get(ncvar.units, ncvar.units)
                if len(src) == 4:
                    sources[dst] = src

        nctime = nc["time"]
        data["time"] = num2pydate(nctime[:], units=nctime.units)

        data["pressure"] = np.tile(data["pressure"], (len(data["time"]), 1))

        data["height"] = calc_geometric_height(nc["HGTS"][:])
        units["height"] = "m"
        sources["height"] = "HGTS converted from gpm to m"

        data["sfc_height"] = calc_geometric_height(nc["SHGT"][:])
        units["sfc_height"] = "m"
        sources["sfc_height"] = "SHGT converted from gpm to m"

        history = nc.history.splitlines()

        return Model(
            GDAS1,
            location,
            data,
            units=units,
            sources=sources,
            comments=comments,
            history=history,
        )


GDAS1 = ModelType(
    id="gdas1",
    full_name="Global Data Assimilation System (GDAS1)",
    short_name="GDAS1",
)



import datetime
from dataclasses import dataclass
from os import PathLike

import netCDF4
import numpy as np
import numpy.typing as npt
from numpy import ma

from model_munger.grid import Grid
from model_munger.level import Level, LevelType
from model_munger.model import ModelType
from model_munger.utils import M_TO_KM, slerp
from model_munger.version import __version__ as model_munger_version


@dataclass
class FixedLocation:
    id: str
    name: str
    latitude: float
    longitude: float


@dataclass
class MobileLocation:
    id: str
    name: str
    time: list[datetime.datetime]
    latitude: list[float]
    longitude: list[float]


@dataclass
class RawModel:
    location: FixedLocation | MobileLocation
    model: ModelType
    data: dict[str, npt.NDArray]
    dimensions: dict[str, tuple[str, ...]]
    attributes: dict[str, dict[str, str]]
    history: str | None = None


def write_netcdf(raw: RawModel, filename: str | PathLike) -> None:
    with netCDF4.Dataset(filename, "w", format="NETCDF4_CLASSIC") as nc:
        nc.Conventions = "CF-1.8"
        nc.title = f"{raw.model.short_name} single-site output over {raw.location.name}"
        nc.location = raw.location.name
        nc.source = raw.model.full_name
        nc.model_munger_version = model_munger_version
        if raw.history is not None:
            nc.history = raw.history

        for key in raw.data:
            values = raw.data[key]
            data_type = values.dtype.str[1:]
            fill_value = (
                netCDF4.default_fillvals[data_type] if ma.is_masked(values) else None
            )
            for dimension, size in zip(raw.dimensions[key], values.shape, strict=True):
                if dimension not in nc.dimensions:
                    nc.createDimension(dimension, size)
            ncvar = nc.createVariable(
                key,
                data_type,
                raw.dimensions[key],
                zlib=True,
                fill_value=fill_value,
            )
            if key in raw.attributes:
                for attr, value in raw.attributes[key].items():
                    setattr(ncvar, attr, value)
            ncvar[:] = values


class Extractor:
    def __init__(
        self,
        time: list[datetime.datetime],
        locations: list[FixedLocation | MobileLocation],
        model: ModelType,
        history: str | None = None,
    ) -> None:
        self.time = time
        self.forecast_time = ma.masked_all(len(self.time), dtype=np.float32)
        self.time_unit = datetime.timedelta(hours=1)
        self.locations = locations
        self.model = model
        self.history = history
        self.idx: npt.NDArray | None = None
        self.lat: npt.NDArray | None = None
        self.lon: npt.NDArray | None = None
        self.res: npt.NDArray | None = None
        self.time_data: dict[tuple[str, int], npt.NDArray] = {}
        self.time_level_data: dict[tuple[str, int, int], npt.NDArray] = {}
        self.time_soil_data: dict[tuple[str, int, int], npt.NDArray] = {}
        self.level_data: dict[tuple[str, int], npt.NDArray] = {}
        self.attributes = {
            "forecast_time": {
                "long_name": "Time since initialization of forecast",
                "units": "hours",
            },
            "latitude": {
                "long_name": "Latitude of model gridpoint",
                "standard_name": "latitude",
                "units": "degree_north",
            },
            "longitude": {
                "long_name": "Longitude of model gridpoint",
                "standard_name": "longitude",
                "units": "degree_east",
            },
            "horizontal_resolution": {
                "long_name": "Horizontal resolution of model",
                "units": "km",
            },
        }
        self.dimensions: dict[str, tuple[str, ...]] = {
            "time": ("time",),
            "forecast_time": ("time",),
            "latitude": ("time",),
            "longitude": ("time",),
            "horizontal_resolution": ("time",),
        }
        self.is_pressure: bool | None = None

    def _set_grid(self, grid: Grid) -> None:
        shape = (len(self.locations), len(self.time))
        self.idx = np.empty(shape, dtype=np.intp)
        self.lat = np.empty(shape, dtype=np.float32)
        self.lon = np.empty(shape, dtype=np.float32)
        self.res = np.empty(shape, dtype=np.float32)
        for loc_ind, loc in enumerate(self.locations):
            lat: float | list[float]
            lon: float | list[float]
            if isinstance(loc, MobileLocation):
                lat, lon = [], []
                for time in self.time:
                    tlat, tlon = slerp(time, loc.time, loc.latitude, loc.longitude)
                    lat.append(tlat)
                    lon.append(tlon)
            else:
                lat, lon = loc.latitude, loc.longitude
            (
                self.idx[loc_ind],
                self.lat[loc_ind],
                self.lon[loc_ind],
                self.res[loc_ind],
            ) = grid.find_closest(lat, lon)

    def add_level(self, level: Level) -> None:
        self.attributes[level.variable] = level.attributes

        if self.idx is None:
            self._set_grid(level.grid)
        time_idx = self.time.index(level.time)
        idx = self.idx[:, time_idx]  # type: ignore[index]

        if level.forecast_time is not None:
            self.forecast_time[time_idx] = level.forecast_time / self.time_unit

        if level.kind == LevelType.SURFACE:
            times = (
                list(range(len(self.time)))
                if level.forecast_time is None
                else [time_idx]
            )
            values = level.values[idx]
            for t in times:
                self.time_data[(level.variable, t)] = values
            self.dimensions[level.variable] = ("time",)
        elif level.kind in (LevelType.MODEL, LevelType.PRESSURE):
            is_pressure = level.kind == LevelType.PRESSURE
            if self.is_pressure is None:
                self.is_pressure = is_pressure
            elif self.is_pressure != is_pressure:
                msg = "Cannot have both pressure and model levels"
                raise ValueError(msg)
            if level.forecast_time is not None:
                self.time_level_data[(level.variable, time_idx, level.level_no)] = (
                    level.values[idx]
                )
                self.dimensions[level.variable] = ("time", "level")
            else:
                self.level_data[(level.variable, level.level_no)] = level.values[idx]
                self.dimensions[level.variable] = ("level",)
        elif level.kind == LevelType.SOIL:
            self.time_soil_data[(level.variable, time_idx, level.level_no)] = (
                level.values[idx]
            )
            self.dimensions[level.variable] = ("time", "soil_level")
        else:
            msg = f"Invalid level type: {level.kind}"
            raise RuntimeError(msg)

    def extract_profiles(self) -> list[RawModel]:
        if self.idx is None:
            return []
        time_vars = set()
        time_level_vars = set()
        time_soil_vars = set()
        level_vars = set()
        uniq_level = set()
        uniq_soil = set()
        for var, _time_ind in self.time_data:
            time_vars.add(var)
        for var, _time_ind, level in self.time_level_data:
            time_level_vars.add(var)
            uniq_level.add(level)
        for var, _time_ind, level in self.time_soil_data:
            time_soil_vars.add(var)
            uniq_soil.add(level)
        for var, _level in self.level_data:
            level_vars.add(var)
        model_levels = sorted(uniq_level, reverse=True)
        soil_levels = sorted(uniq_soil)
        n_time = len(self.time)
        n_level = len(model_levels)
        n_soil = len(soil_levels)
        n_location = len(self.locations)

        start_time = self.time[0]
        time = [(t - start_time) / self.time_unit for t in self.time]
        self.attributes["time"] = {
            "long_name": "Time UTC",
            "standard_name": "time",
            "units": f"hours since {start_time:%Y-%m-%d %H:%M:%S} +00:00",
            "axis": "T",
            "calendar": "standard",
        }
        common_data = {
            "time": np.array(time, dtype=np.float32),
            "forecast_time": self.forecast_time,
        }
        hres = np.round(self.res * M_TO_KM)  # type: ignore[operator]

        if self.is_pressure:
            common_data["pressure"] = np.array(model_levels, dtype=np.float32)
            self.dimensions["pressure"] = ("level",)
            self.attributes["pressure"] = {
                "long_name": "Pressure",
                "standard_name": "air_pressure",
                "units": "Pa",
            }
        else:
            common_data["model_level"] = np.array(model_levels, dtype=np.int16)
            self.dimensions["model_level"] = ("level",)
            self.attributes["model_level"] = {
                "long_name": "Model level",
                "standard_name": "model_level_number",
                "units": "1",
            }

        time_data = {
            var: ma.masked_all((n_location, n_time), dtype=np.float32)
            for var in time_vars
        }
        for (var, time_ind), values in self.time_data.items():
            time_data[var][:, time_ind] = values

        time_level_data = {
            var: ma.masked_all((n_location, n_time, n_level), dtype=np.float32)
            for var in time_level_vars
        }
        for (var, time_ind, level), values in self.time_level_data.items():
            z = model_levels.index(level)
            time_level_data[var][:, time_ind, z] = values

        time_soil_data = {
            var: ma.masked_all((n_location, n_time, n_soil), dtype=np.float32)
            for var in time_soil_vars
        }
        for (var, time_ind, level), values in self.time_soil_data.items():
            z = soil_levels.index(level)
            time_soil_data[var][:, time_ind, z] = values

        level_data = {
            var: ma.masked_all((n_location, n_level), dtype=np.float32)
            for var in level_vars
        }
        for (var, level), values in self.level_data.items():
            z = model_levels.index(level)
            level_data[var][:, z] = values

        now = datetime.datetime.now(datetime.timezone.utc)
        history = (
            f"{now:%Y-%m-%d %H:%M:%S} +00:00 - {self.history} "
            f"using model-munger v{model_munger_version}"
        )

        raws = []
        for loc_idx, loc in enumerate(self.locations):
            merged = _merge_dicts(
                common_data,
                {key: values[loc_idx] for key, values in time_data.items()},
                {key: values[loc_idx] for key, values in time_level_data.items()},
                {key: values[loc_idx] for key, values in time_soil_data.items()},
                {key: values[loc_idx] for key, values in level_data.items()},
                {
                    "latitude": self.lat[loc_idx],  # type: ignore[index]
                    "longitude": self.lon[loc_idx],  # type: ignore[index]
                    "horizontal_resolution": hres[loc_idx],
                },
            )
            raw = RawModel(
                location=loc,
                model=self.model,
                data=merged,
                dimensions=self.dimensions,
                attributes=self.attributes,
                history=history,
            )
            raws.append(raw)

        return raws


def _merge_dicts(*dicts: dict) -> dict:
    output: dict = {}
    for d in dicts:
        conflicted = output.keys() & d.keys()
        if conflicted:
            msg = "Conflicting keys: " + ", ".join(conflicted)
            raise KeyError(msg)
        output.update(d)
    return output



import numpy as np
from numpy import ma

from model_munger.metadata import ATTRIBUTES
from model_munger.model import Model


def merge_models(models: list[Model]) -> Model:
    times = []
    ftimes = []
    mindices = []
    tindices = []
    for i, model in enumerate(models):
        time = model.data["time"]
        ftime = model.data["forecast_time"]
        times.append(time)
        ftimes.append(ftime)
        mindices.append(np.full_like(time, i))
        tindices.append(np.arange(len(time)))
    time = np.concatenate(times)
    ftime = np.concatenate(ftimes)
    mindex = np.concatenate(mindices)
    tindex = np.concatenate(tindices)

    # Sort by forecast time.
    findex = np.argsort(ftime)
    time = time[findex]
    ftime = ftime[findex]
    mindex = mindex[findex]
    tindex = tindex[findex]

    # Sort and find unique times while keeping the smallest forecast time.
    _utime, uindex = np.unique(time, return_index=True)
    mindex = mindex[uindex]
    tindex = tindex[uindex]

    # Initialize merged model with common scalar and 1D data.
    data = {
        key: values
        for key, values in models[0].data.items()
        if key != "time" and "time" not in ATTRIBUTES[key].dimensions
    }

    # Combine 2D data.
    for key in models[0].data:
        if key in data:
            continue
        values = []
        for i, t in zip(mindex, tindex, strict=True):
            if key in models[i].data:
                values.append(models[i].data[key][t : t + 1])
            else:
                shape = (1, *models[0].data[key].shape[1:])
                values.append(ma.masked_all(shape))
        data[key] = ma.concatenate(values)

    used_models = [models[i] for i in np.unique(mindex)]
    return Model(
        models[0].type,
        models[0].location,
        data,
        sources=models[0].sources,
        comments=models[0].comments,
        history=[line for m in used_models for line in m.history],
    )



import datetime

import numpy as np
import numpy.typing as npt
import pytest
from numpy import ma
from numpy.testing import assert_allclose, assert_array_equal

from model_munger import utils


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (np.array([1, 2, 3]), [1, 2, 3]),
        (ma.array([]), []),
        (ma.array([1, 2, 3]), [1, 2, 3]),
        (ma.array([1, 2, 3], mask=[False, False, False]), [1, 2, 3]),
        (
            ma.array([1, 2, 3], mask=[True, False, False]),
            ma.array([1, 2, 3], mask=[True, False, False]),
        ),
        (ma.array([1, 2, 3], mask=[False, True, False]), [1, 1, 3]),
        (ma.array([1, 2, 3], mask=[False, True, True]), [1, 1, 1]),
        (
            ma.array([1, 2, 3], mask=[True, True, True]),
            ma.array([1, 2, 3], mask=[True, True, True]),
        ),
    ],
)
def test_fill_masked(test_input: npt.NDArray, expected: npt.NDArray) -> None:
    actual = utils.ffill(test_input)
    assert_array_equal(actual, expected)
    assert_array_equal(ma.getmaskarray(actual), ma.getmaskarray(expected))


@pytest.mark.parametrize(
    "time,expected",
    [
        # Before
        (datetime.datetime(2025, 9, 10, 10, 0), (59.0, 23.0)),
        # Exact
        (datetime.datetime(2025, 9, 10, 11, 0), (59.0, 23.0)),
        (datetime.datetime(2025, 9, 10, 12, 0), (60.0, 24.0)),
        (datetime.datetime(2025, 9, 10, 13, 0), (61.0, 25.0)),
        # Between
        (datetime.datetime(2025, 9, 10, 11, 30), (59.500954, 23.492592)),
        (datetime.datetime(2025, 9, 10, 12, 30), (60.500935, 24.492287)),
        # After
        (datetime.datetime(2025, 9, 10, 14, 0), (61.0, 25.0)),
    ],
)
def test_slerp(
    time: datetime.datetime,
    expected: tuple[npt.NDArray, npt.NDArray],
) -> None:
    times = [
        datetime.datetime(2025, 9, 10, 11, 0),
        datetime.datetime(2025, 9, 10, 12, 0),
        datetime.datetime(2025, 9, 10, 13, 0),
    ]
    latitudes = [59.0, 60.0, 61.0]
    longitudes = [23.0, 24.0, 25.0]
    actual = utils.slerp(time, times, latitudes, longitudes)
    assert_allclose(actual, expected)


@pytest.mark.parametrize(
    "time,expected",
    [
        (datetime.datetime(2025, 9, 10, 12, 0), (0.0, 179.0)),
        (datetime.datetime(2025, 9, 10, 12, 15), (0.0, 179.5)),
        (datetime.datetime(2025, 9, 10, 12, 30), (0.0, 180.0)),
        (datetime.datetime(2025, 9, 10, 12, 45), (0.0, -179.5)),
        (datetime.datetime(2025, 9, 10, 13, 0), (0.0, -179.0)),
    ],
)
def test_slerp_antimeridian(
    time: datetime.datetime,
    expected: tuple[npt.NDArray, npt.NDArray],
) -> None:
    times = [
        datetime.datetime(2025, 9, 10, 12, 0),
        datetime.datetime(2025, 9, 10, 13, 0),
    ]
    latitudes = [0.0, 0.0]
    longitudes = [179.0, -179.0]
    actual = utils.slerp(time, times, latitudes, longitudes)
    assert_allclose(actual, expected)



from dataclasses import dataclass
from typing import Literal


@dataclass(frozen=True)
class Metadata:
    units: str
    long_name: str
    dimensions: tuple[str, ...] = ()
    standard_name: str | None = None
    comment: str | None = None
    axis: Literal["X", "Y", "Z", "T"] | None = None
    positive: Literal["up", "down"] | None = None


ATTRIBUTES = {
    "latitude": Metadata(
        units="degree_north",
        long_name="Latitude of model gridpoint",
        standard_name="latitude",
        dimensions=("time",),
    ),
    "longitude": Metadata(
        units="degree_east",
        long_name="Longitude of model gridpoint",
        standard_name="longitude",
        dimensions=("time",),
    ),
    "forecast_time": Metadata(
        units="hours",
        long_name="Time since initialization of forecast",
        comment=(
            "For each profile in the file this variable contains the time\n"
            "elapsed since the initialization time of the forecast from which\n"
            "it was taken. Note that the profiles in this file may be taken\n"
            "from more than one forecast."
        ),
        dimensions=("time",),
    ),
    "model_level": Metadata(
        units="1",
        long_name="Model level",
        standard_name="model_level_number",
        axis="Z",
        positive="down",
        dimensions=("level",),
    ),
    "flux_level": Metadata(
        units="1",
        long_name="Flux level",
        standard_name="flux_level_number",
        axis="Z",
        positive="down",
        dimensions=("level",),
    ),
    "height": Metadata(
        units="m",
        long_name="Height above ground",
        standard_name="height",
        dimensions=("time", "level"),
    ),
    "horizontal_resolution": Metadata(
        long_name="Horizontal resolution of model",
        units="km",
        dimensions=("time",),
    ),
    "pressure": Metadata(
        units="Pa",
        long_name="Pressure",
        standard_name="air_pressure",
        dimensions=("time", "level"),
    ),
    "temperature": Metadata(
        units="K",
        long_name="Temperature",
        standard_name="air_temperature",
        dimensions=("time", "level"),
    ),
    "uwind": Metadata(
        units="m s-1",
        long_name="Zonal wind",
        standard_name="eastward_wind",
        dimensions=("time", "level"),
    ),
    "vwind": Metadata(
        units="m s-1",
        long_name="Meridional wind",
        standard_name="northward_wind",
        dimensions=("time", "level"),
    ),
    "wwind": Metadata(
        units="m s-1",
        long_name="Vertical wind",
        standard_name="upward_air_velocity",
        dimensions=("time", "level"),
    ),
    "omega": Metadata(
        units="Pa s-1",
        long_name="Vertical wind in pressure coordinates",
        standard_name="omega",
        dimensions=("time", "level"),
    ),
    "rh": Metadata(
        units="1",
        long_name="Relative humidity",
        standard_name="relative_humidity",
        dimensions=("time", "level"),
    ),
    "q": Metadata(
        units="1",
        long_name="Specific humidity",
        standard_name="specific_humidity",
        dimensions=("time", "level"),
    ),
    "ql": Metadata(
        units="1",
        long_name="Gridbox-mean liquid water mixing ratio",
        standard_name="mass_fraction_of_cloud_liquid_water_in_air",
        dimensions=("time", "level"),
    ),
    "qi": Metadata(
        units="1",
        long_name="Gridbox-mean ice water mixing ratio",
        standard_name="mass_fraction_of_cloud_ice_in_air",
        dimensions=("time", "level"),
    ),
    "qs": Metadata(
        units="1",
        long_name="Gridbox-mean snow mixing ratio",
        standard_name="mass_fraction_of_snow_in_air",
        dimensions=("time", "level"),
    ),
    "qr": Metadata(
        units="1",
        long_name="Gridbox-mean rain mixing ratio",
        standard_name="mass_fraction_of_rain_in_air",
        dimensions=("time", "level"),
    ),
    "qg": Metadata(
        units="1",
        long_name="Gridbox-mean graupel mixing ratio",
        standard_name="mass_fraction_of_graupel_in_air",
        dimensions=("time", "level"),
    ),
    "qh": Metadata(
        units="1",
        long_name="Gridbox-mean hail mixing ratio",
        standard_name="mass_fraction_of_hail_in_air",
        dimensions=("time", "level"),
    ),
    "NL": Metadata(
        units="kg-1",
        long_name="Liquid water droplet number concentration",
        dimensions=("time", "level"),
    ),
    "NI": Metadata(
        units="kg-1",
        long_name="Ice particle number concentration",
        dimensions=("time", "level"),
    ),
    "NS": Metadata(
        units="kg-1",
        long_name="Snow particle number concentration",
        dimensions=("time", "level"),
    ),
    "NR": Metadata(
        units="kg-1",
        long_name="Rain drop number concentration",
        dimensions=("time", "level"),
    ),
    "NG": Metadata(
        units="kg-1",
        long_name="Graupel number concentration",
        dimensions=("time", "level"),
    ),
    "NH": Metadata(
        units="kg-1",
        long_name="Hail number concentration",
        dimensions=("time", "level"),
    ),
    "NI_act": Metadata(
        units="kg-1",
        long_name="Activated ice nuclei number concentration",
        dimensions=("time", "level"),
    ),
    "q_diag": Metadata(
        units="1",
        long_name="Total specific humidity (diagnostic)",
        dimensions=("time", "level"),
    ),
    "ql_diag": Metadata(
        units="1",
        long_name="Total specific liquid water (diagnostic)",
        dimensions=("time", "level"),
    ),
    "qi_diag": Metadata(
        units="1",
        long_name="Total specific ice water (diagnostic)",
        dimensions=("time", "level"),
    ),
    "ls_cloud_fraction": Metadata(
        units="1",
        long_name="Large scale cloud fraction",
        standard_name="large_scale_cloud_area_fraction",
        dimensions=("time", "level"),
    ),
    "conv_cloud_fraction": Metadata(
        units="1",
        long_name="Convective cloud fraction",
        standard_name="convective_cloud_area_fraction",
        dimensions=("time", "level"),
    ),
    "cloud_fraction": Metadata(
        units="1",
        long_name="Cloud fraction",
        standard_name="cloud_area_fraction",
        dimensions=("time", "level"),
    ),
    "tke": Metadata(
        units="m2 s-2",
        long_name="Turbulent kinetic energy",
        standard_name="specific_turbulent_kinetic_energy_of_air",
        dimensions=("time", "level"),
    ),
    "turb_mom_coeff": Metadata(
        units="m2 s-1",
        long_name="Turbulent diffusion coefficients for momentum",
        dimensions=("time", "level"),
    ),
    "turb_heat_coeff": Metadata(
        units="m2 s-1",
        long_name="Turbulent diffusion coefficients for heat",
        dimensions=("time", "level"),
    ),
    "ozone": Metadata(
        units="1",
        long_name="Grid-box mean ozone mixing ratio",
        dimensions=("time", "level"),
    ),
    "mass_flux_up": Metadata(
        units="kg m-2 s-1",
        long_name="Mean updraft mass flux",
        dimensions=("time", "level"),
    ),
    "mass_flux_down": Metadata(
        units="kg m-2 s-1",
        long_name="Mean downdraft mass flux",
        dimensions=("time", "level"),
    ),
    "detrainment_rate_up": Metadata(
        units="kg m-3 s-1",
        long_name="Mean updraft detrainment rate",
        dimensions=("time", "level"),
    ),
    "detrainment_rate_down": Metadata(
        units="kg m-3 s-1",
        long_name="Mean downdraft detrainment rate",
        dimensions=("time", "level"),
    ),
    "total_precipitation_flux": Metadata(
        units="kg m-2 s-1",
        long_name="Mean total precipitation flux",
        dimensions=("time", "level"),
    ),
    "flx_net_sw": Metadata(
        units="W m-2",
        long_name="Net shortwave flux",
        standard_name="net_downward_shortwave_flux_in_air",
        dimensions=("time", "level"),
    ),
    "flx_net_lw": Metadata(
        units="W m-2",
        long_name="Net longwave flux",
        standard_name="net_downward_longwave_flux_in_air",
        dimensions=("time", "level"),
    ),
    "flx_down_sw": Metadata(
        units="W m-2",
        long_name="Downwelling shortwave flux",
        standard_name="downwelling_shortwave_flux_in_air",
        dimensions=("time", "level"),
    ),
    "flx_down_lw": Metadata(
        units="W m-2",
        long_name="Downwelling longwave flux",
        standard_name="downwelling_longwave_flux_in_air",
        dimensions=("time", "level"),
    ),
    "flx_down_sw_cs": Metadata(
        units="W m-2",
        long_name="Downwelling clear sky shortwave flux",
        dimensions=("time", "level"),
    ),
    "flx_down_lw_cs": Metadata(
        units="W m-2",
        long_name="Downwelling clear sky longwave flux",
        dimensions=("time", "level"),
    ),
    "flx_down_sens_heat": Metadata(
        units="W m-2",
        long_name="Sensible heat flux",
        dimensions=("time", "level"),
    ),
    "flx_turb_moist": Metadata(
        units="kg m-2 s-1",
        long_name="Turbulent moisture heat flux",
        dimensions=("time", "level"),
    ),
    "flx_ls_rain": Metadata(
        units="kg m-2 s-1",
        long_name="Large-scale rainfall flux",
        standard_name="large_scale_rainfall_flux",
        dimensions=("time", "level"),
    ),
    "flx_ls_snow": Metadata(
        units="kg m-2 s-1",
        long_name="Large-scale snowfall flux",
        standard_name="large_scale_snowfall_flux",
        dimensions=("time", "level"),
    ),
    "flx_graupel_rain": Metadata(
        units="kg m-2 s-1",
        long_name="Large-scale graupel flux",
        standard_name="large_scale_graupel_flux",
        dimensions=("time", "level"),
    ),
    "flx_conv_rain": Metadata(
        units="kg m-2 s-1",
        long_name="Convective rainfall flux",
        standard_name="convective_rainfall_flux",
        dimensions=("time", "level"),
    ),
    "flx_conv_snow": Metadata(
        units="kg m-2 s-1",
        long_name="Convective snowfall flux",
        standard_name="convective_snowfall_flux",
        dimensions=("time", "level"),
    ),
    "flx_turb_mom_u": Metadata(
        units="kg m-1 s-2",
        long_name="Zonal turbulent momentum flux",
        standard_name="downward_eastward_momentum_flux_in_air",
        dimensions=("time", "level"),
    ),
    "flx_turb_mom_v": Metadata(
        units="kg m-1 s-2",
        long_name="Meridional turbulent momentum flux",
        standard_name="downward_northward_momentum_flux_in_air",
        dimensions=("time", "level"),
    ),
    "sfc_pressure": Metadata(
        units="Pa",
        long_name="Surface pressure",
        standard_name="surface_air_pressure",
        dimensions=("time",),
    ),
    "sfc_pressure_amsl": Metadata(
        long_name="Surface pressure at mean sea level",
        units="Pa",
        dimensions=("time",),
    ),
    "sfc_geopotential": Metadata(
        units="m2 s-2",
        long_name="Geopotential",
        standard_name="geopotential",
        dimensions=("time",),
    ),
    "sfc_height": Metadata(
        units="m",
        long_name="Surface height",
        standard_name="height_above_mean_sea_level",
        dimensions=("time",),
    ),
    "sfc_net_sw": Metadata(
        units="W m-2",
        long_name="Surface net downward shortwave flux",
        standard_name="surface_net_downward_shortwave_flux",
        dimensions=("time",),
    ),
    "sfc_net_sw_cs": Metadata(
        units="W m-2",
        long_name="Surface net clear sky downward shortwave flux",
        dimensions=("time",),
    ),
    "sfc_net_lw": Metadata(
        units="W m-2",
        long_name="Surface net downward longwave flux",
        standard_name="surface_net_downward_longwave_flux",
        dimensions=("time",),
    ),
    "sfc_net_lw_cs": Metadata(
        units="W m-2",
        long_name="Clear sky net downward longwave flux",
        dimensions=("time",),
    ),
    "sfc_net_lat_heat_flx": Metadata(
        units="W m-2",
        long_name="Net latent heat flux at the surface",
        dimensions=("time",),
    ),
    "sfc_up_sw": Metadata(
        units="W m-2",
        long_name="Surface upwelling shortwave flux",
        standard_name="surface_upwelling_shortwave_flux",
        dimensions=("time",),
    ),
    "sfc_up_sw_diffuse": Metadata(
        units="W m-2",
        long_name="Surface diffuse upwelling shortwave flux",
        dimensions=("time",),
    ),
    "sfc_up_lw": Metadata(
        units="W m-2",
        long_name="Surface upwelling longwave flux",
        standard_name="surface_upwelling_longwave_flux",
        dimensions=("time",),
    ),
    "sfc_up_lw_cs": Metadata(
        units="W m-2",
        long_name="Clear sky upwelling longwave flux",
        standard_name="surface_upwelling_longwave_flux_in_air_assuming_clear_sky",
        dimensions=("time",),
    ),
    "sfc_down_sw": Metadata(
        units="W m-2",
        long_name="Surface downwelling shortwave flux",
        standard_name="surface_downwelling_shortwave_flux",
        dimensions=("time",),
    ),
    "sfc_down_sw_diffuse": Metadata(
        units="W m-2",
        long_name="Surface diffuse downwelling shortwave flux",
        standard_name="surface_diffuse_downwelling_shortwave_flux_in_air",
        dimensions=("time",),
    ),
    "sfc_down_sw_direct": Metadata(
        units="W m-2",
        long_name="Surface direct downwelling shortwave flux",
        standard_name="surface_direct_downwelling_shortwave_flux_in_air",
        dimensions=("time",),
    ),
    "sfc_down_sw_cs": Metadata(
        units="W m-2",
        long_name="Surface clear sky downwelling shortwave flux",
        standard_name="surface_downwelling_shortwave_flux_in_air_assuming_clear_sky",
        dimensions=("time",),
    ),
    "sfc_down_sw_direct_normal": Metadata(
        units="W m-2",
        long_name="Surface direct normal downwelling shortwave flux",
        dimensions=("time",),
    ),
    "sfc_down_lw": Metadata(
        units="W m-2",
        long_name="Surface downwelling longwave flux",
        standard_name="surface_downwelling_longwave_flux",
        dimensions=("time",),
    ),
    "sfc_down_lw_cs": Metadata(
        units="W m-2",
        long_name="Surface clear sky downwelling longwave flux",
        standard_name="surface_downwelling_longwave_flux_in_air_assuming_clear_sky",
        dimensions=("time",),
    ),
    "toa_net_sw": Metadata(
        units="W m-2",
        long_name="Top of atmosphere net downward shortwave flux",
        dimensions=("time",),
    ),
    "toa_net_sw_cs": Metadata(
        units="W m-2",
        long_name="Top of atmosphere net clear sky downward shortwave flux",
        dimensions=("time",),
    ),
    "toa_net_lw": Metadata(
        units="W m-2",
        long_name="Top of atmosphere net downward longwave flux",
        dimensions=("time",),
    ),
    "toa_net_lw_cs": Metadata(
        units="W m-2",
        long_name="Top of atmosphere net clear sky downward longwave flux",
        dimensions=("time",),
    ),
    "toa_down_sw": Metadata(
        units="W m-2",
        long_name="Top of atmosphere downwelling shortwave flux",
        dimensions=("time",),
    ),
    "toa_up_sw": Metadata(
        units="W m-2",
        long_name="Top of atmosphere upwelling shortwave flux",
        dimensions=("time",),
    ),
    "toa_up_lw": Metadata(
        units="W m-2",
        long_name="Top of atmosphere upwelling longwave flux",
        dimensions=("time",),
    ),
    "toa_up_lw_cs": Metadata(
        units="W m-2",
        long_name="Top of atmosphere upwelling clear sky longwave flux",
        dimensions=("time",),
    ),
    "toa_brightness_temperature": Metadata(
        units="K",
        long_name="Top of atmosphere brightness temperature",
        standard_name="toa_brightness_temperature",
        dimensions=("time",),
    ),
    "sfc_down_lat_heat_flx": Metadata(
        units="W m-2",
        long_name="Latent heat flux",
        standard_name="surface_downward_latent_heat_flux",
        dimensions=("time",),
    ),
    "sfc_down_sens_heat_flx": Metadata(
        units="W m-2",
        long_name="Sensible heat flux",
        standard_name="surface_downward_sensible_heat_flux",
        dimensions=("time",),
    ),
    "sfc_ls_rain": Metadata(
        units="kg m-2",
        long_name="Large-scale rainfall amount",
        standard_name="large_scale_rainfall_amount",
        dimensions=("time",),
    ),
    "sfc_ls_snow": Metadata(
        units="kg m-2",
        long_name="Large-scale snowfall amount",
        standard_name="large_scale_snowfall_amount",
        dimensions=("time",),
    ),
    "sfc_ls_graupel": Metadata(
        units="kg m-2",
        long_name="Large-scale graupel amount",
        standard_name="stratiform_graupel_fall_amount",
        dimensions=("time",),
    ),
    "sfc_ls_hail": Metadata(
        units="kg m-2",
        long_name="Large-scale hail amount",
        standard_name="hail_fall_amount",
        dimensions=("time",),
    ),
    "sfc_ls_rainrate": Metadata(
        units="kg m-2 s-1",
        long_name="Large-scale rainfall intensity",
        dimensions=("time",),
    ),
    "sfc_ls_snowrate": Metadata(
        units="kg m-2 s-1",
        long_name="Large-scale snowfall intensity",
        dimensions=("time",),
    ),
    "sfc_ls_graupelrate": Metadata(
        units="kg m-2 s-1",
        long_name="Large-scale graupel intensity",
        dimensions=("time",),
    ),
    "sfc_conv_rain": Metadata(
        units="kg m-2",
        long_name="Convective rainfall amount",
        standard_name="convective_rainfall_amount",
        dimensions=("time",),
    ),
    "sfc_conv_snow": Metadata(
        units="kg m-2",
        long_name="Convective snowfall amount",
        standard_name="convective_snowfall_amount",
        dimensions=("time",),
    ),
    "sfc_total_rain": Metadata(
        units="kg m-2",
        long_name="Total rainfall amount",
        dimensions=("time",),
    ),
    "sfc_total_snow": Metadata(
        units="kg m-2",
        long_name="Total snowfall amount",
        dimensions=("time",),
    ),
    "sfc_ls_precip_fraction": Metadata(
        units="1",
        long_name="Large-scale precipitation fraction",
        dimensions=("time",),
    ),
    "sfc_cloud_fraction": Metadata(
        units="1",
        long_name="Surface total cloud fraction",
        dimensions=("time",),
    ),
    "sfc_cloud_fraction_low": Metadata(
        units="1",
        long_name="Surface cloud fraction (low clouds)",
        dimensions=("time",),
    ),
    "sfc_cloud_fraction_medium": Metadata(
        units="1",
        long_name="Surface cloud fraction (mid-level clouds)",
        dimensions=("time",),
    ),
    "sfc_cloud_fraction_high": Metadata(
        units="1",
        long_name="Surface cloud fraction (high clouds)",
        dimensions=("time",),
    ),
    "sfc_conv_cloud_fraction": Metadata(
        units="1",
        long_name="Surface convective cloud fraction",
        dimensions=("time",),
    ),
    "sfc_bl_height": Metadata(
        units="m",
        long_name="Boundary layer height",
        dimensions=("time",),
    ),
    "sfc_bl_dissipation": Metadata(
        units="J m-2",
        long_name="Boundary layer height",
        dimensions=("time",),
    ),
    "sfc_cloud_base_height": Metadata(
        units="m",
        long_name="Cloud base height",
        dimensions=("time",),
    ),
    "sfc_albedo": Metadata(
        units="1",
        long_name="Surface albedo",
        standard_name="surface_albedo",
        dimensions=("time",),
    ),
    "sfc_albedo_snow": Metadata(
        units="1",
        long_name="Surface albedo of snow",
        dimensions=("time",),
    ),
    "sfc_albedo_lw_direct": Metadata(
        units="1",
        long_name="Surface albedo (longwave direct)",
        dimensions=("time",),
    ),
    "sfc_albedo_lw_diffuse": Metadata(
        units="1",
        long_name="Surface albedo (longwave diffuse)",
        dimensions=("time",),
    ),
    "sfc_albedo_sw_direct": Metadata(
        units="1",
        long_name="Surface albedo (shortwavewave direct)",
        dimensions=("time",),
    ),
    "sfc_albedo_sw_diffuse": Metadata(
        units="1",
        long_name="Surface albedo (shortwave diffuse)",
        dimensions=("time",),
    ),
    "sfc_temp": Metadata(
        units="K",
        long_name="Surface temperature",
        dimensions=("time",),
    ),
    "sfc_skin_temp": Metadata(
        units="K",
        long_name="Surface skin temperature",
        dimensions=("time",),
    ),
    "sfc_temp_2m": Metadata(
        units="K",
        long_name="Temperature at 2m",
        dimensions=("time",),
    ),
    "sfc_dewpoint_temp_2m": Metadata(
        units="K",
        long_name="Dew point temperature at 2m",
        dimensions=("time",),
    ),
    "sfc_rh_2m": Metadata(
        units="1",
        long_name="Relative humidity at 2m",
        dimensions=("time",),
    ),
    "sfc_q": Metadata(
        units="1",
        long_name="Surface specific humidity",
        standard_name="surface_specific_humidity",
        dimensions=("time",),
    ),
    "sfc_q_2m": Metadata(
        units="1",
        long_name="Specific humidity at 2m",
        dimensions=("time",),
    ),
    "sfc_rough_mom": Metadata(
        units="m",
        long_name="Surface roughness for momentum",
        dimensions=("time",),
    ),
    "sfc_rough_heat": Metadata(
        units="m",
        long_name="Surface roughness for heat",
        dimensions=("time",),
    ),
    "sfc_rough_oro": Metadata(
        units="m",
        long_name="Surface roughness for orography",
        dimensions=("time",),
    ),
    "sfc_friction_velocity": Metadata(
        units="m s-1",
        long_name="Surface friction velocity",
        dimensions=("time",),
    ),
    "sfc_roughness_length": Metadata(
        units="m",
        long_name="Surface roughness length",
        dimensions=("time",),
    ),
    "sfc_drag_coefficient": Metadata(
        units="1",
        long_name="Surface drag coefficient",
        dimensions=("time",),
    ),
    "sfc_temp_snow": Metadata(
        units="K",
        long_name="Surface snow temperature",
        dimensions=("time",),
    ),
    "sfc_snow_density": Metadata(
        units="kg m-3",
        long_name="Surface snow density",
        dimensions=("time",),
    ),
    "sfc_emissivity": Metadata(
        units="1",
        long_name="Surface emissivity",
        dimensions=("time",),
    ),
    "sfc_dissipation": Metadata(
        units="W n-2",
        long_name="Surface dissipation",
        dimensions=("time",),
    ),
    "sfc_soil_moisture": Metadata(
        units="1",
        long_name="Surface soil moisture content",
        dimensions=("time",),
    ),
    "sfc_wind_u_10m": Metadata(
        long_name="Zonal wind at 10 m",
        units="m s-1",
        dimensions=("time",),
    ),
    "sfc_wind_v_10m": Metadata(
        long_name="Meridional wind at 10 m",
        units="m s-1",
        dimensions=("time",),
    ),
    "sfc_wind_gust_10m": Metadata(
        long_name="Wind gust at 10 m",
        units="m s-1",
        dimensions=("time",),
    ),
    "sfc_uwind_gust_10m": Metadata(
        long_name="Zonal wind gust at 10 m",
        units="m s-1",
        dimensions=("time",),
    ),
    "sfc_wwind_gust_10m": Metadata(
        long_name="Meridional wind gust at 10 m",
        units="m s-1",
        dimensions=("time",),
    ),
    "sfc_turb_mom_u": Metadata(
        units="kg m-1 s-2",
        long_name="Surface zonal turbulent momentum flux",
        dimensions=("time",),
    ),
    "sfc_turb_mom_v": Metadata(
        units="kg m-1 s-2",
        long_name="Surface meridional turbulent momentum flux",
        dimensions=("time",),
    ),
    "sfc_eastward_stress": Metadata(
        units="Pa",
        long_name="Surface zonal stress",
        standard_name="surface_downward_eastward_stress",
        dimensions=("time",),
    ),
    "sfc_northward_stress": Metadata(
        units="Pa",
        long_name="Surface meridional stress",
        standard_name="surface_downward_northward_stress",
        dimensions=("time",),
    ),
    "sfc_bt": Metadata(
        units="K",
        long_name="Surface brightness temperature",
        dimensions=("time",),
    ),
    "sfc_global_rad": Metadata(
        units="kg m-2 s-1",
        long_name="Surface global radiation",
        dimensions=("time",),
    ),
    "sfc_lw_heating": Metadata(
        units="kg m-2 s-1",
        long_name="Surface longwave heating",
        dimensions=("time",),
    ),
    "sfc_solar_heating": Metadata(
        units="kg m-2 s-1",
        long_name="Surface solar heating",
        dimensions=("time",),
    ),
    "sfc_conv_heat": Metadata(
        units="kg m-2 s-1",
        long_name="Surface convective heating",
        dimensions=("time",),
    ),
    "sfc_conv_moist": Metadata(
        units="kg m-2 s-1",
        long_name="Surface convective moistening",
        dimensions=("time",),
    ),
    "sfc_vert_diff_accel": Metadata(
        units="kg m-2 s-1",
        long_name="Surface vertical diffusion of acceleration",
        dimensions=("time",),
    ),
    "sfc_vert_diff_moist": Metadata(
        units="kg m-2 s-1",
        long_name="Surface vertical diffusion of moisture",
        dimensions=("time",),
    ),
    "sfc_vert_diff_heat": Metadata(
        units="kg m-2 s-1",
        long_name="Surface vertical diffusion of heat",
        dimensions=("time",),
    ),
    "sfc_visibility": Metadata(
        units="m",
        long_name="Surface visibility",
        dimensions=("time",),
    ),
    "sfc_weg_snow": Metadata(
        units="m",
        long_name="Surface water equivalent snow depth",
        dimensions=("time",),
    ),
    "sfc_qs": Metadata(
        units="1",
        long_name="Gridbox-mean snow mixing ratio at surface",
        dimensions=("time",),
    ),
    "sfc_qg": Metadata(
        units="1",
        long_name="Gridbox-mean graupel mixing ratio at surface",
        dimensions=("time",),
    ),
    "total_column_water_vapour": Metadata(
        units="kg m-2",
        long_name="Total column water vapour",
        dimensions=("time",),
    ),
    "total_column_water": Metadata(
        units="kg m-2",
        long_name="Total column water",
        dimensions=("time",),
    ),
    "total_column_liquid": Metadata(
        units="kg m-2",
        long_name="Total column liquid water",
        dimensions=("time",),
    ),
    "total_column_sc_liquid": Metadata(
        units="kg m-2",
        long_name="Total column supercooled liquid water",
        dimensions=("time",),
    ),
    "soil_depth": Metadata(
        units="m",
        long_name="Depth below ground",
        standard_name="depth",
        dimensions=("time", "soil_level"),
    ),
    "soil_temperature": Metadata(
        units="K",
        long_name="Soil temperature",
        dimensions=("time", "soil_level"),
    ),
    "soil_moisture": Metadata(
        units="m3 m-3",
        long_name="Soil moisture content",
        dimensions=("time", "soil_level"),
    ),
    "sfc_land_cover": Metadata(
        units="1",
        long_name="Land cover",
        standard_name="land_area_fraction",
        dimensions=("time",),
    ),
    "sfc_cape": Metadata(
        units="J kg-1",
        long_name="Convective available potential energy (CAPE)",
        standard_name="atmosphere_convective_available_potential_energy_wrt_surface",
        dimensions=("time",),
    ),
    "sfc_cin": Metadata(
        units="J kg-1",
        long_name="Convective inhibition (CIN)",
        standard_name="atmosphere_convective_inhibition_wrt_surface",
        dimensions=("time",),
    ),
    "standard_lifted_index": Metadata(
        units="K",
        long_name="Standard lifted index",
        dimensions=("time",),
    ),
    "best_4layer_lifted_index": Metadata(
        units="K",
        long_name="Best 4-layer lifted index",
        dimensions=("time",),
    ),
    "sfc_categorical_snow": Metadata(
        units="1",
        long_name="Categorical snow",
        dimensions=("time",),
    ),
    "sfc_categorical_ice": Metadata(
        units="1",
        long_name="Categorical ice",
        dimensions=("time",),
    ),
    "sfc_categorical_freezing_rain": Metadata(
        units="1",
        long_name="Categorical freezing rain",
        dimensions=("time",),
    ),
    "sfc_categorical_rain": Metadata(
        units="1",
        long_name="Categorical rain",
        dimensions=("time",),
    ),
}



import datetime
from typing import Final

import numpy as np
import numpy.typing as npt
from numpy import ma
from scipy.spatial import geometric_slerp

EARTH_RADIUS: Final = 6_371_229
"Radius of the Earth (m) as assumed in ECMWF IFS"

G: Final = 9.80665
"Earth's gravitational acceleration (m s-2)"

MW_RATIO: Final = 0.62198
"Ratio of the molecular weight of water vapor to dry air"

T0: Final = 273.16
"Triple point of water (K)"

HPA_TO_PA: Final = 100
"Multiplicative conversion factor from hehtopascal to pascal"

M_TO_KM: Final = 1e-3
"Multiplicative conversion factor from meter to kilometer"


def calc_geometric_height(height: npt.NDArray) -> npt.NDArray:
    """Convert geopotential height to geometric height.

    Args:
        height: Geopotential height (m)

    Returns:
        Geometric height (m)

    References:
        ECMWF (2023). ERA5: compute pressure and geopotential on model levels,
        geopotential height and geometric height. https://confluence.ecmwf.int/x/JJh0CQ
    """
    return EARTH_RADIUS * height / (EARTH_RADIUS - height)


def calc_vertical_wind(
    height: npt.NDArray,
    sfc_pressure: npt.NDArray,
    pressure: npt.NDArray,
    omega: npt.NDArray,
) -> npt.NDArray:
    """Convert vertical wind from pressure to cartesian coordinates.

    Args:
        height: Height above ground (m)
        sfc_pressure: Surface pressure (Pa)
        pressure: Pressure (Pa)
        omega: Vertical wind (Pa s-1)

    Returns:
        Vertical wind (m s-1)
    """
    dz = np.diff(height, prepend=0, axis=1)
    dp = np.diff(pressure, prepend=sfc_pressure[:, np.newaxis], axis=1)
    return omega * dz / dp


def calc_saturated_vapor_pressure(temperature: npt.NDArray) -> npt.NDArray:
    """Calculate saturation vapor pressure over liquid or ice.

    Based on the given temperature, the saturated vapor pressure is calculated
    over liquid above freezing and over ice below freezing using Goff-Gratch
    formulae.

    Args:
        temperature: Temperature (K).

    Returns:
        Saturation vapor pressure (Pa).

    References:
        Vömel, H. (2016). Saturation vapor pressure formulations.
        http://cires1.colorado.edu/~voemel/vp.html
    """
    ratio = T0 / temperature
    inv_ratio = temperature / T0
    liquid = HPA_TO_PA * 10 ** (
        10.79574 * (1 - ratio)
        - 5.02800 * np.log10(inv_ratio)
        + 1.50475e-4 * (1 - 10 ** (-8.2969 * (inv_ratio - 1)))
        + 0.42873e-3 * (10 ** (4.76955 * (1 - ratio)) - 1)
        + 0.78614
    )
    ice = HPA_TO_PA * 10 ** (
        -9.09718 * (ratio - 1)
        - 3.56654 * np.log10(ratio)
        + 0.876793 * (1 - inv_ratio)
        + np.log10(6.1071)
    )
    return np.where(temperature < T0, ice, liquid)


def ffill(values: npt.NDArray) -> npt.NDArray:
    """Forward-fills masked values in a 1D NumPy array.

    Args:
        values: Input 1D array, possibly with masked values.

    Returns:
        Array with masked values replaced by the most recent non-masked value.
    """
    mask = ma.getmaskarray(values)
    idx = np.where(mask, 0, np.arange(len(values)))
    np.maximum.accumulate(idx, out=idx)
    return values[idx]


def spherical_to_cartesian(
    latitude: npt.ArrayLike,
    longitude: npt.ArrayLike,
) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
    latitude = np.radians(latitude)
    longitude = np.radians(longitude)
    x = np.cos(latitude) * np.cos(longitude)
    y = np.cos(latitude) * np.sin(longitude)
    z = np.sin(latitude)
    return x, y, z


def cartesian_to_spherical(x: float, y: float, z: float) -> tuple[float, float]:
    hyp = np.hypot(x, y)
    lat = np.degrees(np.arctan2(z, hyp))
    lon = np.degrees(np.arctan2(y, x))
    return lat, lon


def slerp(
    time: datetime.datetime,
    times: list[datetime.datetime],
    latitudes: list[float],
    longitudes: list[float],
) -> tuple[float, float]:
    """Perform spherical linear interpolation.

    Args:
        time: Interpolate point at this time.
        times: Sorted time array for latitudes and longitudes.
        latitudes: Latitudes (degrees).
        longitudes: Longitudes (degrees).

    Returns:
        Latitude and longitude of interpolated point at given time.
    """
    i = np.searchsorted(times, time)  # type: ignore[call-overload]
    if i == 0:
        return latitudes[0], longitudes[0]
    if i == len(times):
        return latitudes[-1], longitudes[-1]
    t = (time - times[i - 1]) / (times[i] - times[i - 1])
    start = spherical_to_cartesian(latitudes[i - 1], longitudes[i - 1])
    end = spherical_to_cartesian(latitudes[i], longitudes[i])
    new_point = geometric_slerp(start, end, t)  # type: ignore[arg-type]
    return cartesian_to_spherical(*new_point)



from os import PathLike

from model_munger.model import Location, Model, ModelType
from model_munger.readers.arpege import _read_lfa2nc


def read_arome(file: str | PathLike, location: Location) -> Model:
    """Read AROME netCDF generated using lfa2nc."""
    return _read_lfa2nc(file, location, AROME)


AROME = ModelType(
    id="arome",
    full_name="Application of Research to Operations at MEsoscale (AROME)",
    short_name="AROME",
)



from model_munger.model import Location, ModelType

SNARK = ModelType(
    id="snark",
    short_name="SNARK",
    full_name="System for Numerical Atmospheric Research and Kinetics (SNARK)",
)
HELSINKI = Location(id="helsinki", name="Helsinki")



import datetime
from dataclasses import dataclass
from enum import Enum

import numpy.typing as npt

from model_munger.grid import Grid


class LevelType(Enum):
    SURFACE = 1
    PRESSURE = 2
    SOIL = 3
    MODEL = 4


@dataclass
class Level:
    kind: LevelType
    level_no: int
    variable: str
    values: npt.NDArray
    time: datetime.datetime
    forecast_time: datetime.timedelta | None
    grid: Grid
    attributes: dict[str, str]



import argparse
import datetime
import sys
from pathlib import Path

from model_munger.cloudnet import api_client, submit_file
from model_munger.download import download_file
from model_munger.extract import (
    Extractor,
    FixedLocation,
    MobileLocation,
    write_netcdf,
)
from model_munger.extractors.ecmwf_open import generate_ecmwf_url, read_ecmwf
from model_munger.extractors.gdas1 import generate_gdas1_url, read_gdas1
from model_munger.readers.ecmwf_open import ECMWF_OPEN
from model_munger.readers.gdas1 import GDAS1


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--date",
        type=parse_date,
        help="Fetch ECMWF open data for this date. Default is today.",
    )
    parser.add_argument(
        "--start",
        type=parse_date,
        help="Fetch ECMWF open data starting from this date. Default is today.",
    )
    parser.add_argument(
        "--stop",
        type=parse_date,
        help="Fetch ECMWF open data until this date. Default is today.",
    )
    parser.add_argument(
        "-r",
        "--runs",
        type=lambda x: [int(y) for y in x.split(",")],
        default=[0],
        help="Comma-separated list of model runs to download.",
    )
    parser.add_argument(
        "--steps",
        type=int,
        default=90,
        help="Maximum time step. Default is 90 hours.",
    )
    parser.add_argument(
        "-s",
        "--sites",
        type=lambda x: x.split(","),
        help="Comma-separated list of Cloudnet sites (e.g. hyytiala) to extract.",
    )
    parser.add_argument(
        "-m",
        "--model",
        choices=["ecmwf-open", "gdas1"],
        help="Which model to download and process.",
        required=True,
    )
    parser.add_argument(
        "--source",
        choices=["ecmwf", "noaa", "aws"],
        help="Where to download ECMWF open data from.",
    )
    parser.add_argument(
        "--submit",
        action="store_true",
        help="Submit files to Cloudnet.",
    )
    parser.add_argument(
        "--no-keep",
        action="store_true",
        help="Don't keep downloaded and processed files.",
    )

    args = parser.parse_args()

    if args.date and (args.start or args.stop):
        parser.error("Cannot use --date with --start and --stop")
    if args.date:
        args.start = args.date
        args.stop = args.date
    else:
        if not args.start:
            args.start = utctoday()
        if not args.stop:
            args.stop = utctoday()
        if args.start > args.stop:
            parser.error("--start should be before --stop")
    del args.date

    if args.sites:
        all_sites = api_client.sites()
        invalid_ids = set(args.sites) - {site.id for site in all_sites}
        if invalid_ids:
            parser.error("Invalid sites: " + ",".join(invalid_ids))
            sys.exit(1)
        sites = [site for site in all_sites if site.id in args.sites]
    else:
        sites = api_client.sites("cloudnet")

    download_dir = Path("data")
    output_dir = Path("output")
    download_dir.mkdir(exist_ok=True)
    output_dir.mkdir(exist_ok=True)

    date = args.start
    while date <= args.stop:
        for run in args.runs:
            locations: list[FixedLocation | MobileLocation] = []
            for site in sites:
                if site.latitude is not None and site.longitude is not None:
                    locations.append(
                        FixedLocation(
                            id=site.id,
                            name=site.human_readable_name,
                            latitude=site.latitude,
                            longitude=site.longitude,
                        ),
                    )
                else:
                    locs = []
                    for day_offset in range(-1, args.steps // 24 + 2):
                        loc_date = date + datetime.timedelta(days=day_offset)
                        locs.extend(api_client.moving_site_locations(site.id, loc_date))
                    locations.append(
                        MobileLocation(
                            id=site.id,
                            name=site.human_readable_name,
                            time=[loc.time for loc in locs],
                            latitude=[loc.latitude for loc in locs],
                            longitude=[loc.longitude for loc in locs],
                        ),
                    )

            if args.model == "ecmwf-open":
                model = ECMWF_OPEN
                history = f"Model run {run:02} UTC extracted from ECMWF open data"
                steps = list(range(0, args.steps + 1, 3))
                start_time = datetime.datetime.combine(
                    date,
                    datetime.time(run),
                    datetime.timezone.utc,
                )
                time = [start_time + datetime.timedelta(hours=step) for step in steps]
                extractor = Extractor(time, locations, model, history)

                date_id = f"{date:%Y%m%d}{run:02}0000"
                source = args.source or "ecmwf"
                for step in steps:
                    url = generate_ecmwf_url(date, run, step, source)
                    path = download_file(url, download_dir)
                    for level in read_ecmwf(path):
                        extractor.add_level(level)
                    if args.no_keep:
                        path.unlink()
            elif args.model == "gdas1":
                model = GDAS1
                source = args.source or "noaa"
                url, revalidate = generate_gdas1_url(date, source)
                filename = url.rsplit("/", maxsplit=1)[-1]
                history = f"GDAS1 data on {date:%Y-%m-%d} extracted from {filename}"
                time = [
                    datetime.datetime.combine(
                        date,
                        datetime.time(hour),
                        datetime.timezone.utc,
                    )
                    for hour in range(0, 24, 3)
                ]
                extractor = Extractor(time, locations, model, history)

                date_id = f"{date:%Y%m%d}"
                path = download_file(url, download_dir, revalidate=revalidate)
                for level in read_gdas1(path):
                    if level.time.date() < date:
                        continue
                    if level.time.date() > date:
                        break
                    extractor.add_level(level)

                next_date = date + datetime.timedelta(days=1)
                next_url, _next_revalidate = generate_gdas1_url(next_date, source)
                if args.no_keep and (date == args.stop or url != next_url):
                    path.unlink()

            for raw in extractor.extract_profiles():
                outfile = f"{date_id}_{raw.location.id}_{raw.model.id}.nc"
                outpath = output_dir / outfile
                print(outpath)
                write_netcdf(raw, outpath)
                if args.submit:
                    submit_file(outpath, raw.location, date, raw.model)
                if args.no_keep:
                    outpath.unlink()

        date += datetime.timedelta(days=1)


def utctoday() -> datetime.date:
    return datetime.datetime.now(datetime.timezone.utc).date()


def parse_date(value: str) -> datetime.date:
    if value == "today":
        return utctoday()
    if value == "yesterday":
        return utctoday() - datetime.timedelta(days=1)
    return datetime.date.fromisoformat(value)


if __name__ == "__main__":
    main()



import datetime
from os import PathLike
from pathlib import Path

import netCDF4
import numpy as np
from cftime import num2pydate
from numpy import ma

from model_munger.model import Location, Model, ModelType
from model_munger.utils import G, calc_geometric_height

keymap = {
    "cc": "cloud_fraction",
    "h_soil": "soil_depth",
    "height_f": "height",
    "lsm": "sfc_land_cover",
    "nlev": "model_level",
    "omega": "omega",
    "orog": "sfc_geopotential",
    "pressure_f": "pressure",
    "ps": "sfc_pressure",
    "q": "q",
    "q2m": "sfc_q_2m",
    "q_soil": "soil_moisture",
    "qi": "qi",
    "ql": "ql",
    "snow": "sfc_weg_snow",
    "t": "temperature",
    "t2m": "sfc_temp_2m",
    "t_soil": "soil_temperature",
    "u": "uwind",
    "u10m": "sfc_wind_u_10m",
    "v": "vwind",
    "v10m": "sfc_wind_v_10m",
}

units_map = {
    "0-1": "1",
    "Pa/s": "Pa s-1",
    "Pascal": "Pa",
    "count": "1",
    "deg E": "degree_east",
    "deg N": "degree_north",
    "kg/kg": "1",
    "m, liquid equivalent": "m",
    "m/s": "m s-1",
    "m2/s2": "m2 s-2",
    "m^3/m^3": "m3 m-3",
}


def _read_lfa2nc(
    file: str | PathLike,
    location: Location,
    model_type: ModelType,
) -> Model:
    with netCDF4.Dataset(file) as nc:
        data: dict = {}
        units = {}

        for src, dst in keymap.items():
            var = nc[src]
            values = var[:]
            values = ma.masked_where(values == -999, values)
            if "nlev" in var.dimensions:
                values = np.flip(values, var.dimensions.index("nlev"))
            data[dst] = values
            units[dst] = _normalize_units(var.units)

        # Height in the input files appears to be geometric height above sea
        # level. Convert this to geometric height above ground based on surface
        # geopotential.
        ground = calc_geometric_height(data["sfc_geopotential"] / G)
        data["height"] = data["height"] - ground[:, np.newaxis]

        time = nc["time"]
        if time.units == "seconds":
            date_int = nc["date"][0]
            year, month_day = divmod(date_int, 10000)
            month, day = divmod(month_day, 100)
            epoch = datetime.datetime(year, month, day) + datetime.timedelta(
                seconds=int(nc["second"][0]),
            )
            data["time"] = np.array(
                [epoch + datetime.timedelta(seconds=int(t)) for t in time],
            )
        else:
            data["time"] = num2pydate(time[:], units=time.units)

        data["latitude"] = nc["lat"][0]
        units["latitude"] = _normalize_units(nc["lat"].units)

        data["longitude"] = nc["lon"][0]
        units["longitude"] = _normalize_units(nc["lon"].units)

        data["soil_depth"] = np.tile(data["soil_depth"], (len(data["time"]), 1))

        history = [
            f"{nc.NetCdf_creation_date} - {Path(file).name} created by {nc.creator}",
        ]

        return Model(model_type, location, data, units, history=history)


def read_arpege(file: str | PathLike, location: Location) -> Model:
    """Read ARPEGE netCDF generated using lfa2nc."""
    return _read_lfa2nc(file, location, ARPEGE)


def _normalize_units(units: str) -> str:
    return units_map.get(units, units)


ARPEGE = ModelType(
    id="arpege",
    full_name="Action de Recherche Petite Echelle Grande Echelle (ARPEGE)",
    short_name="ARPEGE",
)



import numpy as np
import numpy.typing as npt

from model_munger.utils import EARTH_RADIUS


class Grid:
    def find_closest(
        self,
        latitudes: npt.ArrayLike,
        longitudes: npt.ArrayLike,
    ) -> tuple[
        npt.NDArray[np.intp],
        npt.NDArray[np.floating],
        npt.NDArray[np.floating],
        npt.NDArray[np.floating],
    ]:
        raise NotImplementedError


class RegularGrid(Grid):
    def __init__(
        self,
        nlat: int,
        nlon: int,
        first_lat: float,
        first_lon: float,
        last_lat: float,
        last_lon: float,
        delta_lat: float,
        delta_lon: float,
    ) -> None:
        self.nlat = nlat
        self.nlon = nlon
        self.first_lat = first_lat
        self.first_lon = first_lon
        self.last_lat = last_lat
        self.last_lon = last_lon
        self.delta_lat = delta_lat
        self.delta_lon = delta_lon

    def find_closest(
        self,
        latitudes: npt.ArrayLike,
        longitudes: npt.ArrayLike,
    ) -> tuple[
        npt.NDArray[np.intp],
        npt.NDArray[np.floating],
        npt.NDArray[np.floating],
        npt.NDArray[np.floating],
    ]:
        """Finds closest grid points to given coordinates.

        Args:
            latitudes: Latitudes (degrees).
            longitudes: Longitudes (degrees).

        Returns:
            Tuple with array indices, latitudes (degrees), longitudes (degrees)
            and horizontal resolutions (m) of the closest grid points.
        """
        latitudes = np.atleast_1d(latitudes)
        longitudes = np.atleast_1d(longitudes)

        i = (
            np.round((latitudes - self.first_lat) / self.delta_lat).astype(np.intp)
            % self.nlat
        )
        j = (
            np.round((longitudes - self.first_lon) / self.delta_lon).astype(np.intp)
            % self.nlon
        )
        closest_lat = self.first_lat + i * self.delta_lat
        closest_lon = self.first_lon + j * self.delta_lon
        closest_lon[closest_lon > 180] -= 360

        res = (
            np.radians(np.abs(self.delta_lon))
            * EARTH_RADIUS
            * np.cos(np.radians(closest_lat))
        )

        idx = np.ravel_multi_index((i, j), (self.nlat, self.nlon))

        return idx, closest_lat, closest_lon, res



import numpy as np

from model_munger.grid import RegularGrid


def test_gdas1_grid() -> None:
    grid = RegularGrid(181, 360, -90.0, 0.0, 90.0, -1.0, 1.0, 1.0)
    idx, lat, lon, res = grid.find_closest(-89.9, 0.1)
    assert idx == 0
    assert lat == -90
    assert lon == 0
    assert np.isclose(res, 6.80897028e-12)


def test_ecmwf_open_grid() -> None:
    grid = RegularGrid(721, 1440, 90.0, 180.0, -90.0, 179.75, -0.25, 0.25)
    idx, lat, lon, res = grid.find_closest(89.9, 179.9)
    assert idx == 0
    assert lat == 90
    assert lon == 180
    assert np.isclose(res, 1.70224257e-12)



import datetime
from pathlib import Path

import netCDF4
import numpy as np
from numpy.testing import assert_array_equal

from model_munger.model import Model

from .common import HELSINKI, SNARK


def test_model_latlon_scalar(tmp_path: Path) -> None:
    time = np.array(
        [
            datetime.datetime(2024, 1, 22, 0, 0, 0) + datetime.timedelta(hours=i)
            for i in range(25)
        ],
    )
    height = np.array([[10, 100, 1000]] * 25, dtype="f4")
    latitude = np.array(60.25, dtype="f4")
    longitude = np.array(25.0, dtype="f4")
    model = Model(
        SNARK,
        HELSINKI,
        {"time": time, "height": height, "latitude": latitude, "longitude": longitude},
    )
    model.screen_forecast_time(0, 13)
    assert model.data["time"].shape == (13,)
    assert model.data["height"].shape == (13, 3)
    assert model.data["latitude"].shape == (13,)
    assert model.data["longitude"].shape == (13,)

    filename = tmp_path / "model.nc"
    model.write_netcdf(filename)
    with netCDF4.Dataset(filename) as nc:
        assert_array_equal(nc["time"][:], np.arange(13, dtype="f4"), strict=True)
        assert_array_equal(nc["latitude"][:], latitude, strict=True)
        assert_array_equal(nc["longitude"][:], longitude, strict=True)


def test_model_latlon_array(tmp_path: Path) -> None:
    time = np.array(
        [
            datetime.datetime(2024, 1, 22, 0, 0, 0) + datetime.timedelta(hours=i)
            for i in range(25)
        ],
    )
    height = np.array([[10, 100, 1000]] * 25, dtype="f4")
    latitude = np.round(np.linspace(60, 61, 25, dtype="f4") / 0.25) * 0.25
    longitude = np.round(np.linspace(25, 26, 25, dtype="f4") / 0.25) * 0.25
    model = Model(
        SNARK,
        HELSINKI,
        {"time": time, "height": height, "latitude": latitude, "longitude": longitude},
    )
    model.screen_forecast_time(0, 13)
    assert model.data["time"].shape == (13,)
    assert model.data["height"].shape == (13, 3)
    assert model.data["latitude"].shape == (13,)
    assert model.data["longitude"].shape == (13,)

    filename = tmp_path / "model.nc"
    model.write_netcdf(filename)
    with netCDF4.Dataset(filename) as nc:
        assert_array_equal(nc["time"][:], np.arange(13, dtype="f4"), strict=True)
        assert_array_equal(nc["latitude"][:], latitude[:13], strict=True)
        assert_array_equal(nc["longitude"][:], longitude[:13], strict=True)



import datetime
import logging
import os.path
import re
from collections.abc import Iterable
from typing import Any, Literal

import numpy as np
import pygrib

from model_munger.grid import RegularGrid
from model_munger.level import Level, LevelType
from model_munger.utils import HPA_TO_PA

SOURCES = {
    "ecmwf": "https://data.ecmwf.int/forecasts",
    "aws": "https://ecmwf-forecasts.s3.eu-central-1.amazonaws.com",
}
_unsupported_levtypes = set()

logger = logging.getLogger(__name__)


def generate_ecmwf_url(
    date: datetime.date,
    run: Literal[0, 6, 12, 18],
    step: int,
    source: str,
) -> str:
    """Generate URL for ECMWF high-resolution forecast model (open data subset).

    Args:
        date: Forecast date (UTC)
        run: Forecast run (0, 6, 12 or 18 UTC hour)
        step: Forecast step (0, 1, 2, ...)
        source: Location from which to download files ("ecmwf" or "aws").

    Returns:
        URL for GRIB files
    """
    date_str = date.strftime("%Y%m%d")
    run_str = str(run).zfill(2)
    stream = "oper" if run in (0, 12) else "scda"
    if source not in SOURCES:
        msg = f"Invalid source: {source}"
        raise ValueError(msg)
    base_url = SOURCES[source]
    filename = f"{date_str}{run_str}0000-{step}h-{stream}-fc.grib2"
    return f"{base_url}/{date_str}/{run_str}z/ifs/0p25/{stream}/{filename}"


def read_ecmwf(filename: str | os.PathLike) -> Iterable[Level]:
    basename = os.path.basename(filename)
    m = re.match(r"^(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)-(\d+)h-", basename)
    if m is None:
        msg = f"Invalid filename: {basename}"
        raise ValueError(msg)
    start_time = datetime.datetime(
        year=int(m[1]),
        month=int(m[2]),
        day=int(m[3]),
        hour=int(m[4]),
        minute=int(m[5]),
        second=int(m[6]),
        tzinfo=datetime.timezone.utc,
    )
    forecast_time = datetime.timedelta(hours=int(m[7]))
    time = start_time + forecast_time
    with pygrib.open(filename) as grbs:
        for grb in grbs:
            level = grb.level
            if grb.levtype == "sfc":
                kind = LevelType.SURFACE
            elif grb.levtype == "pl":
                kind = LevelType.PRESSURE
                if grb.pressureUnits == "hPa":
                    level *= HPA_TO_PA
                elif grb.pressureUnits != "Pa":
                    msg = f"Invalid pressure units: {grb.pressureUnits}"
                    raise ValueError(msg)
            elif grb.levtype == "sol":
                kind = LevelType.SOIL
            else:
                if grb.levtype not in _unsupported_levtypes:
                    logger.warning("Unsupported level type: %s", grb.levtype)
                    _unsupported_levtypes.add(grb.levtype)
                continue
            attributes = {
                "long_name": grb.name,
                "units": grb.units,
                "param_id": grb.paramId,
            }
            if "cfName" in grb.keys() and grb.cfName != "unknown":  # noqa: SIM118
                attributes["standard_name"] = grb.cfName
            time_invariant = grb.shortName in ("z", "slor", "sdor")
            yield Level(
                kind=kind,
                level_no=level,
                variable=grb.cfVarName,
                values=np.ravel(grb.values),
                grid=_make_grid(grb),
                time=time,
                forecast_time=forecast_time if not time_invariant else None,
                attributes=attributes,
            )


def _make_grid(grb: Any) -> RegularGrid:
    if grb.gridType != "regular_ll":
        msg = f"Invalid grid type: {grb.gridType}"
        raise ValueError(msg)
    delta_lat = grb.jDirectionIncrementInDegrees
    if not grb.jScansPositively:
        delta_lat = -delta_lat
    delta_lon = grb.iDirectionIncrementInDegrees
    if grb.iScansNegatively:
        delta_lon = -delta_lon
    return RegularGrid(
        grb.Nj,
        grb.Ni,
        grb.latitudeOfFirstGridPointInDegrees,
        grb.longitudeOfFirstGridPointInDegrees,
        grb.latitudeOfLastGridPointInDegrees,
        grb.longitudeOfLastGridPointInDegrees,
        delta_lat,
        delta_lon,
    )



import datetime
from os import PathLike

import netCDF4
import numpy as np

from model_munger.model import Location, Model, ModelType

sfc_keymap = {
    "CLCH": "sfc_cloud_fraction_high",
    "CLCL": "sfc_cloud_fraction_low",
    "CLCM": "sfc_cloud_fraction_medium",
    "CLCT": "sfc_cloud_fraction",
    "P_SFC": "sfc_pressure",
    "QV_S": "sfc_q",
    "T2M": "sfc_temp_2m",
    "TD2M": "sfc_dewpoint_temp_2m",
    "T_S": "sfc_temp",
    "T_SNOW": "sfc_temp_snow",
    "U10M": "sfc_wind_u_10m",
    "V10M": "sfc_wind_v_10m",
    "VIS": "sfc_visibility",
    "W_SNOW": "sfc_weg_snow",
}

keymap = {
    "CLC": "cloud_fraction",
    "P": "pressure",
    "QC": "ql",
    "QG": "qg",
    "QI": "qi",
    "QR": "qr",
    "QS": "qs",
    "QV": "q",
    "REL_HUM": "rh",
    "T": "temperature",
    "TKE": "tke",
    "U": "uwind",
    "V": "vwind",
    "W": "wwind",
}


units_map = {
    "-": "1",
    "kg kg-1": "1",
    "kg/kg": "1",
    "m H2O": "m",
    "m/s": "m s-1",
    "m^2/s^2": "m2 s-2",
}


class StationMissingError(Exception):
    pass


def read_icon_d2(file: str | PathLike, station_name: str, location: Location) -> Model:
    """Read ICON meteogram file.

    Args:
        file: Path to ICON meteogram file.
        station_name: Station name to extract from the meteogram file.
        location: Location to use for the given station name.

    Returns:
        Model data.

    Raises:
        StationMissingError: If given station doesn't exist.

    References:
        Reinert et al. (2026). DWD Database Reference for the Global and
            Regional ICON and ICON-EPS Forecasting System. Version 2.5.4.
            https://www.dwd.de/SharedDocs/downloads/DE/modelldokumentationen/nwv/icon/icon_dbbeschr_aktuell.pdf

        Prill et al. (2024). Working with the ICON Model.
            https://www.dwd.de/DE/leistungen/nwv_icon_tutorial/pdf_einzelbaende/icon_tutorial2024.pdf
    """
    data = {}
    units = {}
    sources = {}

    with netCDF4.Dataset(file) as nc:
        station_names = netCDF4.chartostring(nc["station_name"][:]).tolist()
        var_names = netCDF4.chartostring(nc["var_name"][:]).tolist()
        var_units = netCDF4.chartostring(nc["var_unit"][:]).tolist()
        var_nlevs = nc["var_nlevs"][:]
        sfcvar_names = netCDF4.chartostring(nc["sfcvar_name"][:]).tolist()
        sfcvar_units = netCDF4.chartostring(nc["sfcvar_unit"][:]).tolist()
        try:
            station_ind = station_names.index(station_name)
        except ValueError as err:
            msg = f"Station {station_name} not in file"
            raise StationMissingError(msg) from err
        for src_ind, src_name in enumerate(var_names):
            if src_name not in keymap:
                continue
            dst_name = keymap[src_name]
            nlevs = var_nlevs[src_ind]
            data[dst_name] = nc["values"][:, :nlevs, src_ind, station_ind][:, ::-1]
            units[keymap[src_name]] = units_map.get(
                var_units[src_ind], var_units[src_ind]
            )
            sources[keymap[src_name]] = src_name
        for src_ind, src_name in enumerate(sfcvar_names):
            if src_name not in sfc_keymap:
                continue
            dst_name = sfc_keymap[src_name]
            data[dst_name] = nc["sfcvalues"][:, src_ind, station_ind]
            units[dst_name] = units_map.get(
                sfcvar_units[src_ind], sfcvar_units[src_ind]
            )
            sources[dst_name] = src_name

        for key in ("wwind", "tke"):
            data[key] = (data[key][:, :-1] + data[key][:, 1:]) / 2
            sources[key] += " interpolated from half to full levels"

        data["time"] = np.array(
            [
                datetime.datetime.strptime(value, "%Y%m%dT%H%M%SZ")
                for value in netCDF4.chartostring(nc["date"][:])
            ]
        )
        n_time = len(data["time"])

        t_ind = var_names.index("T")
        n_level = var_nlevs[t_ind]
        sfc_height = nc["station_hsurf"][station_ind]
        height = nc["heights"][:n_level, t_ind, station_ind][::-1] - sfc_height
        data["height"] = np.tile(height, (n_time, 1))
        units["height"] = "m"
        sources["height"] = "Calculated from HHL - HSURF"

        data["model_level"] = np.arange(n_level, 0, -1, dtype=np.int32)
        units["model_level"] = "1"

        data["latitude"] = np.repeat(nc["station_lat"][station_ind], n_time)
        units["latitude"] = "degree_north"
        sources["latitude"] = "CLAT"

        data["longitude"] = np.repeat(nc["station_lon"][station_ind], n_time)
        units["longitude"] = "degree_east"
        sources["longitude"] = "CLON"

        data["sfc_height"] = np.repeat(sfc_height, n_time)
        units["sfc_height"] = "m"
        sources["sfc_height"] = "HSURF"

    return Model(
        ICON_D2,
        location,
        data,
        units=units,
        sources=sources,
    )


ICON_D2 = ModelType(
    id="icon-d2",
    full_name="ICOsahedral Nonhydrostatic D2 (ICON-D2)",
    short_name="ICON-D2",
)



import datetime

import numpy as np
from numpy import ma
from numpy.testing import assert_array_equal

from model_munger.merge import merge_models
from model_munger.model import Model

from .common import HELSINKI, SNARK


def test_merge() -> None:
    time1 = [
        datetime.datetime(2024, 1, 22, 0, 0, 0) + datetime.timedelta(hours=i)
        for i in range(25)
    ]
    time2 = [
        datetime.datetime(2024, 1, 22, 12, 0, 0) + datetime.timedelta(hours=i)
        for i in range(25)
    ]
    height = [[10, 100, 1000]] * 25
    pressure1 = [[101205, 100129, 89875]] * 25
    pressure2 = [[101212, 100136, 89880]] * 25
    latitude = [60.25] * 25
    longitude = [25.0] * 25
    model1 = Model(
        SNARK,
        HELSINKI,
        {
            "time": np.array(time1),
            "height": np.array(height),
            "pressure": np.array(pressure1),
            "latitude": np.array(latitude),
            "longitude": np.array(longitude),
        },
        history=["model 1 was created"],
    )
    model2 = Model(
        SNARK,
        HELSINKI,
        {
            "time": np.array(time2),
            "height": np.array(height),
            "pressure": np.array(pressure2),
            "latitude": np.array(latitude),
            "longitude": np.array(longitude),
        },
        history=["model 2 was created"],
    )
    merged = merge_models([model1, model2])
    assert merged.type == SNARK
    assert merged.location == HELSINKI
    assert merged.history == ["model 1 was created", "model 2 was created"]
    assert_array_equal(merged.data["latitude"], latitude[:12] + latitude)
    assert_array_equal(merged.data["longitude"], longitude[:12] + longitude)
    assert_array_equal(merged.data["time"], np.array(time1[:12] + time2))
    assert_array_equal(merged.data["pressure"], pressure1[:12] + pressure2)
    assert_array_equal(merged.data["height"], [[10, 100, 1000]] * (12 + 25))
    assert_array_equal(
        merged.data["forecast_time"],
        np.concatenate([np.arange(12), np.arange(25)]),
    )


def test_missing_variable_is_masked() -> None:
    time1 = [
        datetime.datetime(2024, 1, 22, 0, 0, 0) + datetime.timedelta(hours=i)
        for i in range(25)
    ]
    time2 = [
        datetime.datetime(2024, 1, 22, 12, 0, 0) + datetime.timedelta(hours=i)
        for i in range(25)
    ]
    height = np.array([[10, 100, 1000]] * 25)
    pressure1 = np.array([[101205, 100129, 89875]] * 25)
    latitude = [60.25] * 25
    longitude = [25.0] * 25
    model1 = Model(
        SNARK,
        HELSINKI,
        {
            "time": np.array(time1),
            "height": np.array(height),
            "pressure": np.array(pressure1),
            "latitude": np.array(latitude),
            "longitude": np.array(longitude),
        },
        history=["model 1 was created"],
    )
    model2 = Model(
        SNARK,
        HELSINKI,
        {
            "time": np.array(time2),
            "height": np.array(height),
            "latitude": np.array(latitude),
            "longitude": np.array(longitude),
        },
        history=["model 2 was created"],
    )
    merged = merge_models([model1, model2])
    assert merged.type == SNARK
    assert merged.location == HELSINKI
    assert merged.history == ["model 1 was created", "model 2 was created"]
    assert_array_equal(merged.data["latitude"], latitude[:12] + latitude)
    assert_array_equal(merged.data["longitude"], longitude[:12] + longitude)
    assert_array_equal(merged.data["time"], np.array(time1[:12] + time2))
    assert_array_equal(
        merged.data["pressure"],
        ma.concatenate([pressure1[:12], ma.masked_all_like(pressure1)]),
    )
    assert_array_equal(merged.data["height"], [[10, 100, 1000]] * (12 + 25))
    assert_array_equal(
        merged.data["forecast_time"],
        np.concatenate([np.arange(12), np.arange(25)]),
    )



import datetime
import hashlib
import os
from pathlib import Path

import requests
from cloudnet_api_client import APIClient

from model_munger.extract import FixedLocation, MobileLocation
from model_munger.model import ModelType

BASE_URL = os.environ.get("MM_CLOUDNET_URL", "http://localhost:3000").rstrip("/")
AUTH = (
    os.environ.get("MM_CLOUDNET_USERNAME", "admin"),
    os.environ.get("MM_CLOUDNET_PASSWORD", "admin"),
)

api_client = APIClient(BASE_URL + "/api")


def submit_file(
    filename: Path,
    location: FixedLocation | MobileLocation,
    date: datetime.date,
    model: ModelType,
) -> None:
    print(f"Submit {filename.name}")
    md5_hash = hashlib.md5()
    with open(filename, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            md5_hash.update(byte_block)
    checksum = md5_hash.hexdigest()
    payload = {
        "measurementDate": date.isoformat(),
        "model": model.id,
        "filename": filename.name,
        "checksum": checksum,
        "site": location.id,
    }
    res = requests.post(
        f"{BASE_URL}/model-upload/metadata/",
        json=payload,
        auth=AUTH,
        timeout=60,
    )
    if res.status_code == 409:
        return
    res.raise_for_status()
    with open(filename, "rb") as f:
        res = requests.put(
            f"{BASE_URL}/model-upload/data/{checksum}",
            data=f,
            auth=AUTH,
            timeout=60,
        )
        res.raise_for_status()



import datetime
import logging
from dataclasses import dataclass
from datetime import timedelta
from os import PathLike

import netCDF4
import numpy as np
import numpy.typing as npt
from cftime import date2num

from model_munger.metadata import ATTRIBUTES
from model_munger.utils import (
    HPA_TO_PA,
    MW_RATIO,
    G,
    calc_geometric_height,
    calc_saturated_vapor_pressure,
    calc_vertical_wind,
)
from model_munger.version import __version__

logger = logging.getLogger(__name__)


@dataclass
class ModelType:
    id: str
    full_name: str
    short_name: str


@dataclass
class Location:
    id: str
    name: str


class Model:
    def __init__(
        self,
        model_type: ModelType,
        location: Location,
        data: dict[str, npt.NDArray],
        units: dict[str, str] | None = None,
        sources: dict[str, str] | None = None,
        comments: dict[str, str] | None = None,
        history: list[str] | None = None,
    ) -> None:
        self.type = model_type
        self.location = location
        self.history = history if history is not None else []
        self.sources = sources.copy() if sources is not None else {}
        self.comments = comments if comments is not None else {}
        self.data: dict[str, npt.NDArray] = {}
        n_time = len(data["time"])
        for key, raw_value in data.items():
            if key != "time" and key not in ATTRIBUTES:
                logger.info("Unsupported key %s", key)
                continue
            value = raw_value
            if key != "time" and units and units[key] != ATTRIBUTES[key].units:
                value = _convert_units(key, value, units[key], ATTRIBUTES[key].units)
                if key in self.sources:
                    self.sources[key] = (
                        self.sources[key]
                        + f" converted from {units[key]} to {ATTRIBUTES[key].units}"
                    )
            if key in ("latitude", "longitude") and np.ndim(value) == 0:
                self.data[key] = np.repeat(value, n_time)
            else:
                self.data[key] = value
        if "forecast_time" not in self.data:
            init_time = self.data["time"][0]
            hour = timedelta(hours=1)
            self.data["forecast_time"] = np.array(
                [(t - init_time) / hour for t in self.data["time"]],
            )
        if "wwind" not in self.data and "omega" in self.data:
            self.data["wwind"] = calc_vertical_wind(
                self.data["height"],
                self.data["sfc_pressure"],
                self.data["pressure"],
                self.data["omega"],
            )
            self.sources["wwind"] = (
                "Calculated from omega, height and pressure using: w=omega*dz/dp"
            )
        self._calculate_q("q", "rh", "pressure", "temperature")
        self._calculate_q("sfc_q_2m", "sfc_rh_2m", "sfc_pressure", "sfc_temp_2m")
        self._calculate_rh("q", "rh", "pressure", "temperature")
        self._calculate_rh("sfc_q_2m", "sfc_rh_2m", "sfc_pressure", "sfc_temp_2m")
        if "cloud_fraction" in self.data:
            frac = self.data["cloud_fraction"]
            frac[frac < 1e-4] = 0
        if "sfc_height" not in self.data and "sfc_geopotential" in self.data:
            geopotential_height = self.data["sfc_geopotential"] / G
            self.data["sfc_height"] = calc_geometric_height(geopotential_height)
            self.sources["sfc_height"] = "Calculated from sfc_geopotential"

    def _calculate_q(
        self,
        q_key: str,
        rh_key: str,
        pressure_key: str,
        temperature_key: str,
    ) -> None:
        """Calculate specific humidity if missing.

        References:
            Cai, J. (2019). Humidity Measures.
            https://cran.r-project.org/web/packages/humidity/vignettes/humidity-measures.html
        """
        if (
            q_key in self.data
            or rh_key not in self.data
            or pressure_key not in self.data
            or temperature_key not in self.data
        ):
            return
        es = calc_saturated_vapor_pressure(self.data[temperature_key])
        e = self.data[rh_key] * es
        p = self.data[pressure_key]
        self.data[q_key] = (MW_RATIO * e) / (p - (1 - MW_RATIO) * e)
        self.sources[q_key] = (
            f"Calculated from {rh_key}, {temperature_key} and {pressure_key}"
        )

    def _calculate_rh(
        self,
        q_key: str,
        rh_key: str,
        pressure_key: str,
        temperature_key: str,
    ) -> None:
        """Calculate relative humidity if missing.

        References:
            Cai, J. (2019). Humidity Measures.
            https://cran.r-project.org/web/packages/humidity/vignettes/humidity-measures.html
        """
        if (
            rh_key in self.data
            or q_key not in self.data
            or pressure_key not in self.data
            or temperature_key not in self.data
        ):
            return
        p = self.data[pressure_key]
        q = self.data[q_key]
        e = q * p / (MW_RATIO + (1 - MW_RATIO) * q)
        es = calc_saturated_vapor_pressure(self.data[temperature_key])
        self.data[rh_key] = e / es
        self.sources[rh_key] = (
            f"Calculated from {q_key}, {temperature_key} and {pressure_key}"
        )

    def screen_time(self, date: datetime.date) -> None:
        """Screen time to given date (0th and 24th hour included)."""
        next_date = date + datetime.timedelta(days=1)
        t_min = datetime.datetime.combine(date, datetime.time())
        t_max = datetime.datetime.combine(next_date, datetime.time())
        time = self.data["time"]
        mask = (time >= t_min) & (time <= t_max)
        self._screen_data(mask)

    def screen_forecast_time(self, t_min: float, t_max: float) -> None:
        """Screen forecast times between t_min (include) and t_max (exclude)."""
        time = self.data["forecast_time"]
        mask = (time >= t_min) & (time < t_max)
        self._screen_data(mask)

    def _screen_data(self, mask: npt.NDArray[np.bool]) -> None:
        for key, values in self.data.items():
            if key == "time" or "time" in ATTRIBUTES[key].dimensions:
                self.data[key] = values[mask]

    def write_netcdf(self, filename: PathLike | str) -> None:
        with netCDF4.Dataset(filename, "w", format="NETCDF4_CLASSIC") as nc:
            nc.Conventions = "CF-1.8"
            nc.title = (
                f"{self.type.short_name} single-site output over {self.location.name}"
            )
            nc.location = self.location.name
            nc.cloudnet_file_type = "model"
            date = self.data["time"][0].date()
            nc.year = str(date.year)
            nc.month = str(date.month).zfill(2)
            nc.day = str(date.day).zfill(2)
            nc.source = self.type.full_name
            nc.model_munger_version = __version__
            now = datetime.datetime.now(datetime.timezone.utc)
            history = [
                f"{now:%Y-%m-%d %H:%M:%S} +00:00 - "
                f"Cloudnet model file generated using model-munger v{__version__}",
                *self.history,
            ]
            nc.history = "\n".join(history)

            n_time, n_level = self.data["height"].shape
            nc.createDimension("time", n_time)
            nc.createDimension("level", n_level)
            if "soil_depth" in self.data:
                n_time, n_soil = self.data["soil_depth"].shape
                nc.createDimension("soil_level", n_soil)

            ncvar = nc.createVariable("time", "f4", "time", zlib=True)
            ncvar.long_name = "Hours UTC"
            ncvar.units = f"hours since {date:%Y-%m-%d} 00:00:00 +00:00"
            ncvar.standard_name = "time"
            ncvar.axis = "T"
            ncvar.calendar = "standard"
            ncvar[:] = date2num(
                self.data["time"],
                units=ncvar.units,
                calendar=ncvar.calendar,
            )

            for key, meta in ATTRIBUTES.items():
                if key not in self.data:
                    continue
                data_type = self.data[key].dtype.str[1:]
                if data_type == "f8":
                    data_type = "f4"
                fill_value = netCDF4.default_fillvals[data_type]
                values = self.data[key]
                dimensions = meta.dimensions
                if key in ("latitude", "longitude") and np.all(values == values[0]):
                    values = values[0]
                    dimensions = ()
                ncvar = nc.createVariable(
                    key,
                    data_type,
                    dimensions,
                    zlib=True,
                    fill_value=fill_value,
                )
                ncvar.units = meta.units
                ncvar.long_name = meta.long_name
                if meta.standard_name:
                    ncvar.standard_name = meta.standard_name
                if key in self.comments:
                    ncvar.comment = self.comments[key]
                elif meta.comment is not None:
                    ncvar.comment = meta.comment
                if meta.axis:
                    ncvar.axis = meta.axis
                if meta.positive:
                    ncvar.positive = meta.positive
                if key in self.sources:
                    ncvar.source = self.sources[key]
                ncvar[:] = values


def _convert_units(
    key: str,
    values: npt.NDArray,
    units_from: str,
    units_to: str,
) -> npt.NDArray:
    if units_from == "hPa" and units_to == "Pa":
        return values * HPA_TO_PA
    if units_from == "hPa s-1" and units_to == "Pa s-1":
        return values * HPA_TO_PA
    if units_from == "%" and units_to == "1":
        return values / 100
    msg = f"Cannot convert '{key}' from '{units_from}' to '{units_to}'"
    raise ValueError(msg)



import datetime
import os
from collections.abc import Iterator
from typing import BinaryIO

import numpy as np

from model_munger.grid import RegularGrid
from model_munger.level import Level, LevelType
from model_munger.utils import HPA_TO_PA

LONG_NAMES = {
    "PRSS": "Pressure at surface",
    "MSLP": "Pressure reduced to mean sea level",
    "TPP6": "Accumulated precipitation (6 h accumulation)",
    "UMOF": "u-component of momentum flux (3- or 6-h average)",
    "VMOF": "v-component of momentum flux (3- or 6-h average)",
    "SHTF": "Sensible heat net flux at surface (3- or 6-h average)",
    "DSWF": "Downward short wave radiation flux (3- or 6-h average)",
    "RH2M": "Relative Humidity at 2m AGL",
    "U10M": "U-component of wind at 10 m AGL",
    "V10M": "V-component of wind at 10 m AGL",
    "T02M": "Temperature at 2m AGL",
    "TCLD": "Total cloud cover (3- or 6-h average)",
    "SHGT": "Geopotential height",
    "CAPE": "Convective available potential energy",
    "CINH": "Convective inhibition",
    "LISD": "Standard lifted index",
    "LIB4": "Best 4-layer lifted index",
    "PBLH": "Planetary boundary layer height",
    "TMPS": "Temperature at surface",
    "CPP6": "Accumulated convective precipitation (6 h accumulation)",
    "CPPA": "Accumulated convective precipitation (total accumulation)",
    "SOLM": "Volumetric soil moisture content",
    "CSNO": "Categorial snow (yes=1, no=0) (3- or 6-h average)",
    "CICE": "Categorial ice (yes=1, no=0) (3- or 6-h average)",
    "CFZR": "Categorial freezing rain (yes=1, no=0) (3- or 6-h average)",
    "CRAI": "Categorial rain (yes=1, no=0) (3- or 6-h average)",
    "LHTF": "Latent heat net flux at surface (3- or 6-h average)",
    "LCLD": "Low cloud cover (3- or 6-h average)",
    "MCLD": "Middle cloud cover (3- or 6-h average)",
    "HCLD": "High cloud cover (3- or 6-h average)",
    "HGTS": "Geopotential height",
    "TEMP": "Temperature",
    "UWND": "U-component of wind with respect to grid",
    "VWND": "V-component of wind with respect to grid",
    "WWND": "Pressure vertical velocity",
    "RELH": "Relative humidity",
}

UNITS = {
    "PRSS": "hPa",
    "MSLP": "hPa",
    "TPP6": "m",
    "UMOF": "N/m2",
    "VMOF": "N/m2",
    "SHTF": "W/m2",
    "DSWF": "W/m2",
    "RH2M": "%",
    "U10M": "m/s",
    "V10M": "m/s",
    "T02M": "K",
    "TCLD": "%",
    "SHGT": "gpm",
    "CAPE": "J/kg",
    "CINH": "J/kg",
    "LISD": "K",
    "LIB4": "K",
    "PBLH": "m",
    "TMPS": "K",
    "CPP6": "m",
    "CPPA": "m",
    "SOLM": "frac.",
    "LHTF": "W/m2",
    "LCLD": "%",
    "MCLD": "%",
    "HCLD": "%",
    "HGTS": "gpm",
    "TEMP": "K",
    "UWND": "m/s",
    "VWND": "m/s",
    "WWND": "hPa/s",
    "RELH": "%",
}


NOAA_URL = "https://www.ready.noaa.gov/data/archives/gdas1/"
AWS_URL = "https://noaa-oar-arl-hysplit-pds.s3.amazonaws.com/gdas1/"
MONTHS = [
    "jan",
    "feb",
    "mar",
    "apr",
    "may",
    "jun",
    "jul",
    "aug",
    "sep",
    "oct",
    "nov",
    "dec",
]


def generate_gdas1_url(date: datetime.date, source: str) -> tuple[str, bool]:
    """Generate URL for a file in GDAS1 archive.

    Args:
        date: Date (UTC)
        source: Location from which to download files ("ecmwf" or "aws").

    Returns:
        Tuple with URL and boolean that indicates whether the file should be
        revalidated if previously downloaded.
    """
    month = MONTHS[date.month - 1]
    year = date.year % 100
    week = (date.day - 1) // 7 + 1
    filename = f"gdas1.{month}{year:02}.w{week}"
    if source == "noaa":
        today = datetime.datetime.now(datetime.timezone.utc).date()
        current_start = datetime.date(
            today.year,
            today.month,
            7 * ((today.day - 1) // 7) + 1,
        )
        if date >= current_start:
            filename = "current7days"
            revalidate = True
        else:
            revalidate = False
        url = NOAA_URL + filename
    elif source == "aws":
        url = f"{AWS_URL}{date.year}/{filename}"
        revalidate = False
    else:
        msg = f"Invalid source: {source}"
        raise ValueError(msg)
    return url, revalidate


def read_gdas1(filename: str | os.PathLike) -> Iterator[Level]:
    with open(filename, "rb") as f:
        yield from _read(f)


GRID = RegularGrid(181, 360, -90.0, 0.0, 90.0, -1.0, 1.0, 1.0)
GRID_DEF = (
    b"90.0000"
    b"359.000"
    b"1.00000"
    b"1.00000"
    b".000000"
    b".000000"
    b".000000"
    b"1.00000"
    b"1.00000"
    b"-90.000"
    b".000000"
    b".000000"
    b"360"
    b"181"
)


def _read(f: BinaryIO) -> Iterator[Level]:
    while True:
        header = f.read(50)
        if len(header) == 0:
            break
        if header[14:18] != b"INDX":
            msg = "Invalid header"
            raise ValueError(msg)

        header = f.read(108)
        if header[9:99] != GRID_DEF:
            msg = "Unexpected grid definition"
            raise ValueError(msg)
        nx = 360
        ny = 181
        nz = int(header[99:102])
        k_flag = int(header[102:104])
        if k_flag != 2:
            msg = "Expected absolute pressure levels"
            raise ValueError(msg)
        lenh = int(header[104:108])

        heights = []
        for _z in range(nz):
            header = f.read(8)
            height = int(float(header[0:6]) * HPA_TO_PA)
            n_vars = int(header[6:8])
            heights.extend([height] * n_vars)
            f.seek(n_vars * 8, os.SEEK_CUR)

        f.seek(nx * ny - lenh, os.SEEK_CUR)
        for height in heights:
            header = f.read(50)
            year = 2000 + int(header[0:2])
            month = int(header[2:4])
            day = int(header[4:6])
            hour = int(header[6:8])
            forecast_hour = int(header[8:10])
            level = int(header[10:12])
            variable = header[14:18].decode()
            exponent = int(header[18:22])
            precision = float(header[22:36])
            value = float(header[36:50])
            compressed = f.read(nx * ny)
            if forecast_hour == -1:
                continue
            values = np.frombuffer(compressed, dtype=np.uint8).reshape((ny, nx))
            values = (values.astype(np.float32) - 127) / 2 ** (7 - exponent)
            if values[0, 0] != 0:
                msg = "First value not zero"
                raise ValueError(msg)
            values[0, 0] = value
            np.cumsum(values[:, 0], out=values[:, 0])
            np.cumsum(values, axis=1, out=values)
            values[np.abs(values) < precision] = 0

            kind = LevelType.SURFACE if level == 0 else LevelType.PRESSURE
            time = datetime.datetime(
                year,
                month,
                day,
                hour,
                tzinfo=datetime.timezone.utc,
            )
            forecast_time = datetime.timedelta(hours=hour % 6)
            attributes = {"long_name": LONG_NAMES[variable]}
            if variable in UNITS:
                attributes["units"] = UNITS[variable]

            yield Level(
                kind=kind,
                level_no=height,
                variable=variable,
                values=np.ravel(values),
                time=time,
                forecast_time=forecast_time,
                grid=GRID,
                attributes=attributes,
            )



import re
from os import PathLike

import netCDF4
import numpy as np
from cftime import num2pydate

from model_munger.model import Location, Model, ModelType
from model_munger.utils import calc_geometric_height, ffill

keymap = {
    "asn": "sfc_albedo_snow",
    "d2m": "sfc_dewpoint_temp_2m",
    "fg10": "sfc_wind_gust_10m",
    "horizontal_resolution": "horizontal_resolution",
    "latitude": "latitude",
    "longitude": "longitude",
    "lsm": "sfc_land_cover",
    "msl": "sfc_pressure_amsl",
    "pressure": "pressure",
    "q": "q",
    "r": "rh",
    "skt": "sfc_skin_temp",
    "sot": "soil_temperature",
    "sp": "sfc_pressure",
    "t": "temperature",
    "t2m": "sfc_temp_2m",
    "tcw": "total_column_water",
    "tcwv": "total_column_water_vapour",
    "tprate": "sfc_ls_rainrate",
    "u": "uwind",
    "u10": "sfc_wind_u_10m",
    "v": "vwind",
    "v10": "sfc_wind_v_10m",
    "vsw": "soil_moisture",
    "w": "omega",
    "z": "sfc_geopotential",
}

RH_COMMENT = """For temperatures over 0°C (273.15 K) it is calculated for
saturation over water. At temperatures below -23°C it is calculated for
saturation over ice. Between -23°C and 0°C this parameter is calculated by
interpolating between the ice and water values using a quadratic function."""


def read_ecmwf_open(file: str | PathLike, location: Location) -> Model:
    """Read ECMWF open data netCDF generated using model-munger."""
    with netCDF4.Dataset(file) as nc:
        data = {}
        units = {}
        sources = {}
        comments = {"rh": RH_COMMENT}

        for src, dst in keymap.items():
            if src not in nc.variables:
                continue
            var = nc[src]
            data[dst] = var[:]
            units[dst] = _normalize_units(var.units)
            if hasattr(var, "param_id"):
                sources[dst] = f"ECMWF parameter {var.param_id}"

        # Forward-fill values that are available only in the first time step.
        for key in ("sfc_geopotential",):
            if key in data:
                data[key] = ffill(data[key])

        time = nc["time"]
        data["time"] = num2pydate(time[:], units=time.units)

        data["pressure"] = np.tile(data["pressure"], (len(data["time"]), 1))

        if "soil_temperature" in data or "soil_moisture" in data:
            soil_depth = [0.07, 0.21, 0.72, 1.89]
            data["soil_depth"] = np.tile(soil_depth, (len(data["time"]), 1))
            units["soil_depth"] = "m"

        data["height"] = calc_geometric_height(nc["gh"][:])
        units["height"] = "m"
        sources["height"] = (
            f"ECMWF parameter {nc['gh'].param_id} converted from gpm to m"
        )

        history = nc.history.splitlines()

        return Model(ECMWF_OPEN, location, data, units, sources, comments, history)


def _normalize_units(units: str) -> str:
    if units in ("kg kg**-1", "(0 - 1)"):
        return "1"
    return re.sub(r"\*\*(-?\d+)", r"\1", units)


ECMWF_OPEN = ModelType(
    id="ecmwf-open",
    full_name="ECMWF open data",
    short_name="ECMWF open data",
)
